# -*- coding: utf-8 -*-

import numpy

from ffc.uflacs.backends.ufc.generators import *

import ffc.uflacs.language.cnodes as L


# TODO: Make this a feature of dijitso: dijitso show-function modulehash functionname
def extract_function(name, code):
    lines = code.split("\n")
    n = len(lines)
    begin = None
    body = None
    for i in range(n):
        if (name + "(") in lines[i]:
            for j in range(i, n):
                if lines[j] == "{":
                    begin = i
                    body = j
                    break
            break
    if begin is None:
        return "didnt find %s" % (name,)
    end = n
    for i in range(body, n):
        if lines[i] == "}":
            end = i + 1
            break
    sublines = lines[begin:end]
    return '\n'.join(sublines)


def basic_class_properties(classname):
    ir = {
        "classname": classname,
        "constructor": "",
        "constructor_arguments": "",
        "initializer_list": "",
        "destructor": "",
        "members": "",
        "preamble": "",
        "jit": False,
    }
    return ir


def mock_form_ir():
    ir = basic_class_properties("mock_form_classname")
    ir.update({
        "id": 0,
        "prefix": "Foo",
        "signature": "mock form signature",
        "rank": 2,
        "num_coefficients": 3,
        "original_coefficient_position": [0, 2],
    })

    ir.update({
        "create_coordinate_finite_element": ["mock_coordinate_finite_element_classname"],
        "create_coordinate_dofmap": ["mock_coordinate_dofmap_classname"],
        "create_coordinate_mapping": ["mock_coordinate_mapping_classname"],
        "create_finite_element": ["mock_finite_element_classname_%d" % (i,) for i in range(ir["num_coefficients"])],
        "create_dofmap": ["mock_dofmap_classname_%d" % (i,) for i in range(ir["num_coefficients"])],
    })

    # These are the method names in ufc::form that are specialized for each integral type
    template = "max_%s_subdomain_id"
    for i, integral_type in enumerate(ufc_integral_types):
        key = template % integral_type
        ir[key] = i  # just faking some integers

    template = "has_%s_integrals"
    for i, integral_type in enumerate(ufc_integral_types):
        key = template % integral_type
        ir[key] = (i % 2 == 0)  # faking some bools

    template = "create_%s_integral"
    for i, integral_type in enumerate(ufc_integral_types):
        key = template % integral_type
        ir[key] = [key.replace("create_", "") + str(j) for j in range(i)]  # faking list of classnames

    template = "create_default_%s_integral"
    for i, integral_type in enumerate(ufc_integral_types):
        key = template % integral_type
        ir[key] = key.replace("create_", "")  # faking classname

    return ir


def mock_dofmap_ir():
    ir = basic_class_properties("mock_dofmap_classname")
    num_dofs_per_entity = [1, 1, 1]
    entity_dofs = [[(0,), (1,), (2,)],
                   [(3,), (4,), (5,)],
                   [(6,)]]
    entity_closure_dofs = {
        (0, 0): [0],
        (0, 1): [1],
        (0, 2): [2],
        (1, 0): [0, 1],
        (1, 1): [0, 2],
        (1, 2): [1, 2],
        (2, 0): [0, 1, 2],
        }
    ir.update({
        "signature": "mock element signature",
        "geometric_dimension": 3,
        "topological_dimension": 2,
        "global_dimension": ([3, 2, 1], 4),
        "tabulate_dofs": ([[[(0,), (1,), (2,)], [(3,), (4,), (5,)], [(6,)]], None], [7, 1], True, [False, True]),
        "tabulate_facet_dofs": [[0, 1, 2], [1, 2, 3], [0, 2, 3]],
        "tabulate_entity_dofs": (entity_dofs, num_dofs_per_entity),
        "tabulate_entity_closure_dofs": (entity_closure_dofs, entity_dofs, num_dofs_per_entity),
        "needs_mesh_entities": [True, False, True],
        "num_element_dofs": 7,
        "num_entity_dofs": num_dofs_per_entity,
        "num_entity_closure_dofs": [3, 6, 10],
        "num_facet_dofs": 7,
        "num_sub_dofmaps": 3,
        "create_sub_dofmap": ["mock_dofmap_classname_sub_%d" % (i,) for i in range(3)],
    })
    return ir


def mock_evaluate_basis_ir():
    dofs_data = [
        {
            "embedded_degree": 5,
            "num_components": 2,
            "num_expansion_members": 7,
            "coeffs": [list(range(20, 27)), list(range(30, 37))],
            "reference_offset": 3,
            "physical_offset": 5,
        },
        {
            "embedded_degree": 1,
            "num_components": 2,
            "num_expansion_members": 7,
            "coeffs": [list(range(7)), list(range(10, 17))],
            "reference_offset": 3,
            "physical_offset": 5,
        }
    ]
    data = {
        "cellname": "triangle",
        "geometric_dimension": 3,
        "topological_dimension": 2,
        "reference_value_size": 3,
        "physical_value_size": 2,
        "dofs_data": dofs_data,
    }
    return data


def test_mock_evaluate_basis():
    from ffc.uflacs.backends.ufc.evaluatebasis import generate_evaluate_reference_basis
    import ffc.uflacs.language.cnodes as L
    data = mock_evaluate_basis_ir()
    code = generate_evaluate_reference_basis(L, data)
    print(code)


def mock_finite_element_ir():
    ir = basic_class_properties("mock_finite_element_classname")
    ebir = mock_evaluate_basis_ir()
    ir.update({
        "signature": "mock element signature",
        "cell_shape": "mock_cell_shape",
        "geometric_dimension": 3,
        "topological_dimension": 2,
        "degree": 2,
        "family": "Lagrange",
        "value_dimension": (3, 3),
        "reference_value_dimension": (2, 2),
        "space_dimension": 6,
        "tabulate_dof_coordinates": {"gdim": 3, "tdim": 2, "points": [(0.0, 0.0), (0.0, 1.0), (1.0, 0.0)]},
        "evaluate_basis": ebir,
        "evaluate_basis_derivatives": ebir,
        "evaluate_basis_all": ebir,
        "evaluate_basis_derivatives_all": ebir,
        "evaluate_dof": "fixme",
        "evaluate_dofs": "fixme",
        "interpolate_vertex_values": "fixme",
        "num_sub_elements": 3,
        "create_sub_element": ["mock_finite_element_classname_sub_%d" % (i,) for i in range(3)],
    })
    return ir


def mock_integral_ir():
    ir = basic_class_properties("mock_integral_classname")
    ir.update({
        "prefix": "Foo",
        "enabled_coefficients": [True, False, True],
        "tabulate_tensor": "    mock_body_of_tabulate_tensor();",
        "num_cells": 1,
    })
    return ir


def mock_coordinate_mapping_ir():
    ir = basic_class_properties("mock_coordinate_mapping_classname")
    ir.update({
        "signature": "mock_coordinate_mapping_signature",
        "cell_shape": "mock_cell_shape",
        "geometric_dimension": 3,
        "topological_dimension": 2,
        "create_coordinate_finite_element": "mock_coordinate_finite_element_classname",
        "create_coordinate_dofmap": "mock_coordinate_dofmap_classname",
        "scalar_coordinate_finite_element_classname": "mock_scalar_coordinate_finite_element_classname",
        "coordinate_element_degree": 2,
        "num_scalar_coordinate_element_dofs": 5,
        "tables": {"x0": numpy.ones((5,)),
                   "xm": numpy.ones((5,)),
                   "J0": numpy.ones((2, 5)),
                   "Jm": numpy.ones((2, 5)), }
    })
    return ir


def compile_mock_coordinate_mapping():
    ir = mock_coordinate_mapping_ir()
    gen = ufc_coordinate_mapping()
    return gen.generate(L, ir)


def compile_mock_form():
    ir = mock_form_ir()
    gen = ufc_form()
    return gen.generate(L, ir)


def compile_mock_dofmap():
    ir = mock_dofmap_ir()
    gen = ufc_dofmap()
    return gen.generate(L, ir)


def compile_mock_finite_element():
    ir = mock_finite_element_ir()
    gen = ufc_finite_element()
    return gen.generate(L, ir)


def compile_mock_integral(integral_type):
    ir = mock_integral_ir()
    gen = eval("ufc_%s_integral" % integral_type)()
    return gen.generate(L, ir)


def compile_mock_all():
    mocks = [compile_mock_integral(integral_type) for integral_type in ufc_integral_types]
    mocks += [compile_mock_form(), compile_mock_dofmap(), compile_mock_finite_element()]
    return '\n\n'.join(mocks)


def test_mock_coordinate_mapping():
    h, cpp = compile_mock_coordinate_mapping()
    print(h)
    print(cpp)


def test_mock_form():
    h, cpp = compile_mock_form()
    print(h)
    print(cpp)


def test_mock_dofmap():
    h, cpp = compile_mock_dofmap()
    print(h)
    print(cpp)


def test_mock_finite_element():
    h, cpp = compile_mock_finite_element()
    print(h)
    print(cpp)


def test_mock_integral():
    for integral_type in ufc_integral_types:
        h, cpp = compile_mock_integral(integral_type)
        print(h)
        print(cpp)


def test_foo_integral_properties():
    ir = mock_form_ir()
    assert "cell_integral" in ufc_form.create_cell_integral.__doc__
    assert "return" in str(ufc_form().create_cell_integral(L, ir))


def test_mock_extract_function():
    h, cpp = compile_mock_coordinate_mapping()
    name = "compute_reference_coordinates"
    print("/// Extracted", name, ":")
    print("/// begin")
    print(extract_function(name, cpp))
    print("/// end")


def test_debug_by_printing_extracted_function():
    h, cpp = compile_mock_coordinate_mapping()
    # name = "compute_reference_coordinates"
    # name = "compute_physical_coordinates"
    # name = "compute_jacobians"
    name = "compute_jacobian_inverses"
    print("/// Extracted", name, ":")
    print("/// begin")
    print(extract_function(name, cpp))
    print("/// end")


"""
Missing:

finite_element:
evaluate_basis*
evaluate_dof
interpolate_vertex_values
Ok? tabulate_dof_coordinates

integrals:
tabulate_tensor

all:
everything with classnames
"""



# -*- coding: utf-8 -*-
from ffc.quadrature.quadraturerepresentation import compute_integral_ir
from ffc.quadrature.quadratureoptimization import optimize_integral_ir
from ffc.quadrature.quadraturegenerator import generate_integral_code



# -*- coding: utf-8 -*-
# Copyright (C) 2004-2009 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Garth N. Wells, 2006
# Modified by Marie E. Rognes, 2007
# Modified by Kristian B. Oelgaard, 2009
#
# First added:  2004-11-03
# Last changed: 2009-12-21

# Python modules.
import functools
import numpy

# FFC modules.
from ffc.utils import listcopy
from ffc.log import error


def build_indices(dims):
    "Create a list of all index combinations."
    if not dims:
        return [[]]
    ranges = listcopy(dims)
    return functools.reduce(outer_join, ranges, [[]])


def outer_join(a, b):
    """Let a be a list of lists and b a list. We append each element
    of b to each list in a and return the resulting list of lists."""
    outer = []
    for i in range(len(a)):
        for j in range(len(b)):
            outer += [a[i] + [b[j]]]
    return outer


def create_multiindex(indices):
    "Create multiindex for given list of indices."

    # Check that we got all indices correctly
    indices = sorted(indices)
    for (i, index) in enumerate(indices):
        if not i == index.index_id:
            error("Unable to extract all indices.")

    # Get dimensions
    dims = [list(range(len(index.index_range))) for index in indices]

    return MultiIndex(dims)


class MultiIndex:

    """
    A MultiIndex represents a list of indices and holds the following
    data:

        rank    - rank of multiindex
        dims    - a list of dimensions
        indices - a list of all possible multiindex values
    """

    def __init__(self, dims):
        "Create multiindex from given list of ranges"
        self.rank = len(dims)
        self.dims = [len(dim) for dim in dims]
        self.indices = build_indices(dims)
        return

    def __str__(self):
        "Return informal string representation (pretty-print)."
        return "rank = %d dims = %s indices = %s" % (self.rank, str(self.dims), str(self.indices))



# -*- coding: utf-8 -*-

# Copyright (C) 2009-2014 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Anders Logg, 2009, 2015
# Modified by Martin Sandve Alnæs, 2013-2014

"Quadrature representation class."

import numpy
import itertools

# UFL modules
from ufl.cell import num_cell_entities
from ufl.classes import ReferenceGrad, Grad, CellAvg, FacetAvg
from ufl.algorithms import extract_unique_elements, extract_type, extract_elements
from ufl import custom_integral_types

# FFC modules
from ffc.log import error
from ffc.utils import product, insert_nested_dict
from ffc.fiatinterface import create_element
from ffc.fiatinterface import map_facet_points, reference_cell_vertices
from ffc.representationutils import create_quadrature_points_and_weights
from ffc.representationutils import integral_type_to_entity_dim


def _find_element_derivatives(expr, elements, element_replace_map):
    "Find the highest derivatives of given elements in expression."
    # TODO: This is most likely not the best way to get the highest
    #       derivative of an element, but it works!

    # Initialise dictionary of elements and the number of derivatives.
    # (Note that elements are already mapped through the
    # element_replace_map)
    num_derivatives = dict((e, 0) for e in elements)

    # Extract the derivatives from the integral.
    derivatives = set(extract_type(expr, Grad)) | set(extract_type(expr,
                                                                   ReferenceGrad))

    # Loop derivatives and extract multiple derivatives.
    for d in list(derivatives):
        # After UFL has evaluated derivatives, only one element can be
        # found inside any single Grad expression
        elem, = extract_elements(d.ufl_operands[0])
        elem = element_replace_map[elem]
        # Set the number of derivatives to the highest value
        # encountered so far.
        num_derivatives[elem] = max(num_derivatives[elem],
                                    len(extract_type(d, Grad)),
                                    len(extract_type(d, ReferenceGrad)))
    return num_derivatives


def _tabulate_empty_psi_table(tdim, deriv_order, element):
    "Tabulate psi table when there are no points (custom integrals)."

    # All combinations of partial derivatives up to given order
    gdim = tdim  # hack, consider passing gdim variable here
    derivs = [d for d in itertools.product(*(gdim*[list(range(0, deriv_order + 1))]))]
    derivs = [d for d in derivs if sum(d) <= deriv_order]

    # Return empty table
    table = {}
    for d in derivs:
        value_shape = element.value_shape()
        if value_shape == ():
            table[d] = [[]]
        else:
            value_size = product(value_shape)
            table[d] = [[[] for c in range(value_size)]]

    # Let entity be 0 even for non-cells, this is for
    # custom integrals where we don't need tables to
    # contain multiple entitites
    entity = 0
    return {entity: table}


def _map_entity_points(cellname, tdim, points, entity_dim, entity):
    # Not sure if this is useful anywhere else than in _tabulate_psi_table!
    if entity_dim == tdim:
        assert entity == 0
        return points
    elif entity_dim == tdim-1:
        return map_facet_points(points, entity)
    elif entity_dim == 0:
        return (reference_cell_vertices(cellname)[entity],)


def _tabulate_psi_table(integral_type, cellname, tdim,
                        element, deriv_order, points):
    "Tabulate psi table for different integral types."
    # Handle case when list of points is empty
    if points is None:
        return _tabulate_empty_psi_table(tdim, deriv_order, element)

    # Otherwise, call FIAT to tabulate
    entity_dim = integral_type_to_entity_dim(integral_type, tdim)
    num_entities = num_cell_entities[cellname][entity_dim]
    psi_table = {}
    for entity in range(num_entities):
        entity_points = _map_entity_points(cellname, tdim, points, entity_dim, entity)
        psi_table[entity] = element.tabulate(deriv_order, entity_points)
    return psi_table


# MSA: This function is in serious need for some refactoring and
#      splitting up.  Or perhaps I should just add a new
#      implementation for uflacs, but I'd rather not have two versions
#      to maintain.
def tabulate_basis(sorted_integrals, form_data, itg_data):
    "Tabulate the basisfunctions and derivatives."

    # MER: Note to newbies: this code assumes that each integral in
    # the dictionary of sorted_integrals that enters here, has a
    # unique number of quadrature points ...

    # Initialise return values.
    quadrature_rules = {}
    psi_tables = {}
    integrals = {}
    avg_elements = {"cell": [], "facet": []}

    # Get some useful variables in short form
    integral_type = itg_data.integral_type
    cell = itg_data.domain.ufl_cell()
    cellname = cell.cellname()
    tdim = itg_data.domain.topological_dimension()
    entity_dim = integral_type_to_entity_dim(integral_type, tdim)
    num_entities = num_cell_entities[cellname][entity_dim]

    # Create canonical ordering of quadrature rules
    rules = sorted(sorted_integrals.keys())

    # Loop the quadrature points and tabulate the basis values.
    for rule in rules:
        scheme, degree = rule

        # --------- Creating quadrature rule
        # Make quadrature rule and get points and weights.
        (points, weights) = create_quadrature_points_and_weights(integral_type,
                                                                 cell, degree,
                                                                 scheme)

        # The TOTAL number of weights/points
        num_points = None if weights is None else len(weights)

        # Add points and rules to dictionary
        if num_points in quadrature_rules:
            error("This number of points is already present in the weight table: " + repr(quadrature_rules))
        quadrature_rules[num_points] = (weights, points)

        # --------- Store integral
        # Add the integral with the number of points as a key to the
        # return integrals.
        integral = sorted_integrals[rule]
        if num_points in integrals:
            error("This number of points is already present in the integrals: " + repr(integrals))
        integrals[num_points] = integral

        # --------- Analyse UFL elements in integral

        # Get all unique elements in integral.
        ufl_elements = [form_data.element_replace_map[e]
                        for e in extract_unique_elements(integral)]

        # Insert elements for x and J
        domain = integral.ufl_domain()  # FIXME: For all domains to be sure? Better to rewrite though.
        x_element = domain.ufl_coordinate_element()
        if x_element not in ufl_elements:
            if integral_type in custom_integral_types:
                # FIXME: Not yet implemented, in progress
                # warning("Vector elements not yet supported in custom integrals so element for coordinate function x will not be generated.")
                pass
            else:
                ufl_elements.append(x_element)

        # Find all CellAvg and FacetAvg in integrals and extract
        # elements
        for avg, AvgType in (("cell", CellAvg), ("facet", FacetAvg)):
            expressions = extract_type(integral, AvgType)
            avg_elements[avg] = [form_data.element_replace_map[e]
                                 for expr in expressions
                                 for e in extract_unique_elements(expr)]

        # Find the highest number of derivatives needed for each element
        num_derivatives = _find_element_derivatives(integral.integrand(),
                                                    ufl_elements,
                                                    form_data.element_replace_map)
        # Need at least 1 for the Jacobian
        num_derivatives[x_element] = max(num_derivatives.get(x_element, 0), 1)

        # --------- Evaluate FIAT elements in quadrature points and
        # --------- store in tables

        # Add the number of points to the psi tables dictionary
        if num_points in psi_tables:
            error("This number of points is already present in the psi table: " + repr(psi_tables))
        psi_tables[num_points] = {}

        # Loop FIAT elements and tabulate basis as usual.
        for ufl_element in ufl_elements:
            fiat_element = create_element(ufl_element)

            # Tabulate table of basis functions and derivatives in
            # points
            psi_table = _tabulate_psi_table(integral_type, cellname, tdim,
                                            fiat_element,
                                            num_derivatives[ufl_element],
                                            points)

            # Insert table into dictionary based on UFL elements
            # (None=not averaged)
            avg = None
            psi_tables[num_points][ufl_element] = { avg: psi_table }

    # Loop over elements found in CellAvg and tabulate basis averages
    num_points = 1
    for avg in ("cell", "facet"):
        # Doesn't matter if it's exterior or interior
        if avg == "cell":
            avg_integral_type = "cell"
        elif avg == "facet":
            avg_integral_type = "exterior_facet"

        for element in avg_elements[avg]:
            fiat_element = create_element(element)

            # Make quadrature rule and get points and weights.
            (points, weights) = create_quadrature_points_and_weights(avg_integral_type, cell, element.degree(), "default")
            wsum = sum(weights)

            # Tabulate table of basis functions and derivatives in
            # points
            entity_psi_tables = _tabulate_psi_table(avg_integral_type,
                                                    cellname, tdim,
                                                    fiat_element, 0, points)
            rank = len(element.value_shape())

            # Hack, duplicating table with per-cell values for each
            # facet in the case of cell_avg(f) in a facet integral
            if num_entities > len(entity_psi_tables):
                assert len(entity_psi_tables) == 1
                assert avg_integral_type == "cell"
                assert "facet" in integral_type
                v, = sorted(entity_psi_tables.values())
                entity_psi_tables = dict((e, v) for e in range(num_entities))

            for entity, deriv_table in sorted(entity_psi_tables.items()):
                deriv, = sorted(deriv_table.keys())  # Not expecting derivatives of averages
                psi_table = deriv_table[deriv]

                if rank:
                    # Compute numeric integral
                    num_dofs, num_components, num_points = psi_table.shape
                    if num_points != len(weights):
                        error("Weights and table shape does not match.")
                    avg_psi_table = numpy.asarray([[[numpy.dot(psi_table[j, k, :], weights) / wsum]
                                                    for k in range(num_components)]
                                                   for j in range(num_dofs)])
                else:
                    # Compute numeric integral
                    num_dofs, num_points = psi_table.shape
                    if num_points != len(weights):
                        error("Weights and table shape does not match.")
                    avg_psi_table = numpy.asarray([[numpy.dot(psi_table[j, :],
                                                              weights) / wsum] for j in range(num_dofs)])

                # Insert table into dictionary based on UFL elements
                insert_nested_dict(psi_tables, (num_points, element, avg,
                                                entity, deriv), avg_psi_table)

    return (integrals, psi_tables, quadrature_rules)



# -*- coding: utf-8 -*-
"""
This module provides an abstract ErrorControlGenerator class for
generating forms required for goal-oriented error control and a
realization of this: UFLErrorControlGenerator for handling pure UFL
forms.
"""

# Copyright (C) 2010 Marie E. Rognes
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

from ufl import inner, dx, ds, dS, avg, replace, action

__all__ = ["ErrorControlGenerator", "UFLErrorControlGenerator"]


class ErrorControlGenerator:

    def __init__(self, module, F, M, u):
        """
        *Arguments*

            module (Python module)
               The module to use for specific form manipulations
               (typically ufl or dolfin)

            F (tuple or Form)
               tuple of (bilinear, linear) forms or linear form

            M (Form)
               functional or linear form

            u (Coefficient)
              The coefficient considered as the unknown.
        """
        # Store module
        self.module = module

        # Store solution Coefficient/Function
        self.u = u

        # Extract the lhs (bilinear form), rhs (linear form), goal
        # (functional), weak residual (linear form)
        linear_case = (isinstance(F, (tuple, list)) and len(F) == 2)
        if linear_case:
            self.lhs, self.rhs = F
            try:
                self.goal = action(M, u)
            except:
                self.goal = M    # Allow functionals as input as well
            self.weak_residual = self.rhs - action(self.lhs, u)
        else:
            self.lhs = self.module.derivative(F, u)
            self.rhs = F
            self.goal = M
            self.weak_residual = - F

        # At least check that final forms have correct rank
        assert(len(self.lhs.arguments()) == 2)
        assert(len(self.rhs.arguments()) == 1)
        assert(len(self.goal.arguments()) == 0)
        assert(len(self.weak_residual.arguments()) == 1)

        # Get the domain
        self.domain = self.weak_residual.ufl_domain()

        # Store map from identifiers to names for forms and generated
        # coefficients
        self.ec_names = {}

        # Use predefined names for the forms in the primal problem
        self.ec_names[id(self.lhs)] = "lhs"
        self.ec_names[id(self.rhs)] = "rhs"
        self.ec_names[id(self.goal)] = "goal"

        # Initialize other required data
        self.initialize_data()

    def initialize_data(self):
        """
        Initialize specific data
        """
        msg = """ErrorControlGenerator acts as an abstract
        class. Subclasses must overload the initialize_data() method
        and provide a certain set of variables. See
        UFLErrorControlGenerator for an example."""
        raise NotImplementedError(msg)

    def generate_all_error_control_forms(self):
        """
        Utility function for generating all (8) forms required for
        error control in addition to the primal forms
        """
        # Generate dual forms
        (a_star, L_star) = self.dual_forms()

        # Generate forms for computing strong cell residual
        (a_R_T, L_R_T) = self.cell_residual()

        # Generate forms for computing strong facet residuals
        (a_R_dT, L_R_dT) = self.facet_residual()

        # Generate form for computing error estimate
        eta_h = self.error_estimate()

        # Generate form for computing error indicators
        eta_T = self.error_indicators()

        # Paranoid checks added after introduction of multidomain features in ufl:
        for i, form in enumerate((a_star, L_star, eta_h, a_R_T, L_R_T, a_R_dT, L_R_dT, eta_T)):
            assert form.ufl_domain() is not None

        # Return all generated forms in CERTAIN order matching
        # constructor of dolfin/adaptivity/ErrorControl.h
        return (a_star, L_star, eta_h, a_R_T, L_R_T, a_R_dT, L_R_dT, eta_T)

    def primal_forms(self):
        """
        Return primal forms in order (bilinear, linear, functional)
        """
        return self.lhs, self.rhs, self.goal

    def dual_forms(self):
        """
        Generate and return (bilinear, linear) forms defining linear
        dual variational problem
        """
        a_star = self.module.adjoint(self.lhs)
        L_star = self.module.derivative(self.goal, self.u)
        return (a_star, L_star)

    def cell_residual(self):
        """
        Generate and return (bilinear, linear) forms defining linear
        variational problem for the strong cell residual
        """
        # Define trial and test functions for the cell residuals on
        # discontinuous version of primal trial space
        R_T = self.module.TrialFunction(self._dV)
        v = self.module.TestFunction(self._dV)

        # Extract original test function in the weak residual
        v_h = self.weak_residual.arguments()[0]

        # Define forms defining linear variational problem for cell
        # residual
        v_T = self._b_T * v
        a_R_T = inner(v_T, R_T) * dx(self.domain)
        L_R_T = replace(self.weak_residual, {v_h: v_T})

        return (a_R_T, L_R_T)

    def facet_residual(self):
        """
        Generate and return (bilinear, linear) forms defining linear
        variational problem for the strong facet residual(s)
        """
        # Define trial and test functions for the facet residuals on
        # discontinuous version of primal trial space
        R_e = self.module.TrialFunction(self._dV)
        v = self.module.TestFunction(self._dV)

        # Extract original test function in the weak residual
        v_h = self.weak_residual.arguments()[0]

        # Define forms defining linear variational problem for facet
        # residual
        v_e = self._b_e * v
        a_R_dT = ((inner(v_e('+'), R_e('+')) +
                   inner(v_e('-'), R_e('-'))) * dS(self.domain) +
                  inner(v_e, R_e) * ds(self.domain))
        L_R_dT = (replace(self.weak_residual, {v_h: v_e}) -
                  inner(v_e, self._R_T) * dx(self.domain))

        return (a_R_dT, L_R_dT)

    def error_estimate(self):
        """
        Generate and return functional defining error estimate
        """
        # Error estimate defined as r(Ez_h):
        eta_h = action(self.weak_residual, self._Ez_h)
        return eta_h

    def error_indicators(self):
        """
        Generate and return linear form defining error indicators
        """
        # Extract these to increase readability
        R_T = self._R_T
        R_dT = self._R_dT
        z = self._Ez_h
        z_h = self._z_h

        # Define linear form for computing error indicators
        v = self.module.TestFunction(self._DG0)
        eta_T = (v * inner(R_T, z - z_h) * dx(self.domain) +
                 avg(v)*(inner(R_dT('+'), (z - z_h)('+')) +
                         inner(R_dT('-'), (z - z_h)('-'))) * dS(self.domain) +
                 v * inner(R_dT, z - z_h) * ds(self.domain))

        return eta_T


class UFLErrorControlGenerator(ErrorControlGenerator):

    """
    This class provides a realization of ErrorControlGenerator for use
    with pure UFL forms
    """

    def __init__(self, F, M, u):
        """
        *Arguments*

            F (tuple or Form)
               tuple of (bilinear, linear) forms or linear form

            M (Form)
               functional or linear form

            u (Coefficient)
              The coefficient considered as the unknown.
        """

        ErrorControlGenerator.__init__(self, __import__("ufl"), F, M, u)

    def initialize_data(self):
        """
        Extract required objects for defining error control
        forms. This will be stored, reused and in particular named.
        """
        # Developer's note: The UFL-FFC-DOLFIN--PyDOLFIN toolchain for
        # error control is quite fine-tuned. In particular, the order
        # of coefficients in forms is (and almost must be) used for
        # their assignment. This means that the order in which these
        # coefficients are defined matters and should be considered
        # fixed.

        from ufl import FiniteElement, FunctionSpace, Coefficient
        from ufl.algorithms.elementtransformations import tear, increase_order

        # Primal trial element space
        self._V = self.u.ufl_function_space()

        # Extract domain
        domain = self.u.ufl_domain()

        # Primal test space == Dual trial space
        Vhat = self.weak_residual.arguments()[0].ufl_function_space()

        # Discontinuous version of primal trial element space
        self._dV = FunctionSpace(domain, tear(self._V.ufl_element()))

        # Extract geometric dimension
        gdim = domain.geometric_dimension()

        # Coefficient representing improved dual
        E = FunctionSpace(domain, increase_order(Vhat.ufl_element()))
        self._Ez_h = Coefficient(E)
        self.ec_names[id(self._Ez_h)] = "__improved_dual"

        # Coefficient representing cell bubble function
        Belm = FiniteElement("B", domain.ufl_cell(), gdim + 1)
        Bfs = FunctionSpace(domain, Belm)
        self._b_T = Coefficient(Bfs)
        self.ec_names[id(self._b_T)] = "__cell_bubble"

        # Coefficient representing strong cell residual
        self._R_T = Coefficient(self._dV)
        self.ec_names[id(self._R_T)] = "__cell_residual"

        # Coefficient representing cell cone function
        Celm = FiniteElement("DG", domain.ufl_cell(), gdim)
        Cfs = FunctionSpace(domain, Celm)
        self._b_e = Coefficient(Cfs)
        self.ec_names[id(self._b_e)] = "__cell_cone"

        # Coefficient representing strong facet residual
        self._R_dT = Coefficient(self._dV)
        self.ec_names[id(self._R_dT)] = "__facet_residual"

        # Define discrete dual on primal test space
        self._z_h = Coefficient(Vhat)
        self.ec_names[id(self._z_h)] = "__discrete_dual_solution"

        # Piecewise constants for assembling indicators
        self._DG0 = FunctionSpace(domain, FiniteElement("DG",
                                                        domain.ufl_cell(), 0))



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testExpandOperations():
    f0 = FloatValue(-1)
    f1 = FloatValue(2)
    f2 = FloatValue(1)
    sx = Symbol("x", GEO)
    sy = Symbol("y", GEO)
    sz = Symbol("z", GEO)
    s0 = Product([FloatValue(-1), Symbol("x", GEO)])
    s1 = Symbol("y", GEO)
    s2 = Product([FloatValue(5), Symbol("z", IP)])
    s3 = Product([FloatValue(-4), Symbol("z", GEO)])

    # Random variable values
    x = 2.2
    y = -0.2
    z = 1.1

    # Aux. expressions
    P0 = Product([s2, s1])
    P1 = Product([P0, s0])
    P2 = Product([P1, s1, P0])
    P3 = Product([P1, P2])

    S0 = Sum([s2, s1])
    S1 = Sum([S0, s0])
    S2 = Sum([S1, s1, S0])
    S3 = Sum([S1, S2])

    F0 = Fraction(s2, s1)
    F1 = Fraction(F0, s0)
    F2 = Fraction(F1, F0)
    F3 = Fraction(F1, F2)

    # Special fractions
    F4 = Fraction(P0, F0)
    F5 = Fraction(Fraction(s0, P0), P0)
    F6 = Fraction(Fraction(Fraction(s1, s0), Fraction(s1, s2)), Fraction(Fraction(s2, s0), Fraction(s1, s0)))
    F7 = Fraction(s1, Product([s1, Symbol("x", GEO)]))
    F8 = Fraction(Sum([sx, Fraction(sy, sx)]), FloatValue(2))

    F4x = F4.expand()
    F5x = F5.expand()
    F6x = F6.expand()
    F7x = F7.expand()
    F8x = F8.expand()

    assert round(eval(str(F4)) - eval(str(F4x)), 10) == 0.0
    assert round(eval(str(F5)) - eval(str(F5x)), 10) == 0.0
    assert round(eval(str(F6)) - eval(str(F6x)), 10) == 0.0
    assert round(eval(str(F7)) - eval(str(F7x)), 10) == 0.0
    assert round(eval(str(F8)) - eval(str(F8x)), 10) == 0.0

    assert F4.ops() == 5
    assert F4x.ops() == 1
    assert F5.ops() == 6
    assert F5x.ops() == 5
    assert F6.ops() == 9
    assert F6x.ops() == 1
    assert F7.ops() == 2
    assert F7x.ops() == 1
    assert F8.ops() == 3
    assert F8x.ops() == 4

    # Expressions that should be expanded
    e0 = Product([P3, F2])
    e1 = Product([S3, P2])
    e2 = Product([F3, S1])

    e3 = Sum([P3, F2])
    e4 = Sum([S3, P2])
    e5 = Sum([F3, S1])

    e6 = Fraction(P3, F2)
    e7 = Fraction(S3, P2)
    e8 = Fraction(F3, S1)
    e9 = Fraction(S0, s0)

    e0x = e0.expand()
    e1x = e1.expand()
    e2x = e2.expand()
    e3x = e3.expand()
    e4x = e4.expand()
    e5x = e5.expand()
    e6x = e6.expand()
    e7x = e7.expand()
    e8x = e8.expand()
    e9x = e9.expand()

    assert round(eval(str(e0)) - eval(str(e0x)), 10) == 0.0
    assert round(eval(str(e1)) - eval(str(e1x)), 10) == 0.0
    assert round(eval(str(e2)) - eval(str(e2x)), 10) == 0.0
    assert round(eval(str(e3)) - eval(str(e3x)), 10) == 0.0
    assert round(eval(str(e4)) - eval(str(e4x)), 10) == 0.0
    assert round(eval(str(e5)) - eval(str(e5x)), 10) == 0.0
    assert round(eval(str(e6)) - eval(str(e6x)), 10) == 0.0
    assert round(eval(str(e7)) - eval(str(e7x)), 10) == 0.0
    assert round(eval(str(e8)) - eval(str(e8x)), 10) == 0.0
    assert round(eval(str(e9)) - eval(str(e9x)), 10) == 0.0

    assert e0.ops() == 16
    assert e0x.ops() == 8
    assert e1.ops() == 18
    assert e1x.ops() == 23
    assert e2.ops() == 14
    assert e2x.ops() == 9
    assert e3.ops() == 16
    assert e3x.ops() == 11
    assert e4.ops() == 18
    assert e4x.ops() == 12
    assert e5.ops() == 14
    assert e5x.ops() == 6
    assert e6.ops() == 16
    assert e6x.ops() == 10
    assert e7.ops() == 18
    assert e7x.ops() == 17
    assert e8.ops() == 14
    assert e8x.ops() == 8
    assert e9.ops() == 3
    assert e9x.ops() == 4

    # More expressions (from old expand tests)
    PF = Product([F0, F1])
    E0 = Product([s1, f0, S1])
    E1 = Sum([P0, E0])
    E2 = Fraction(Sum([Product([f1])]), f2)
    E3 = Sum([F0, F0])
    E4 = Product([Sum([Product([sx, Sum([sy, Product([Sum([sy, Product([sy, sz]), sy])]), sy])]),
                       Product([sx, Sum([Product([sy, sz]), sy])])])])
    P4 = Product([s1, Sum([s0, s1])])
    P5 = Product([s0, E0])
    P6 = Product([s1])
    S4 = Sum([s1])

    # Create 'real' term that caused me trouble
    P00 = Product([Symbol("Jinv_00", GEO)] * 2)
    P01 = Product([Symbol("Jinv_01", GEO)] * 2)
    P20 = Product([Symbol("Jinv_00", GEO),
                   Product([f1, Symbol("Jinv_20", GEO)])])
    P21 = Product([Symbol("Jinv_01", GEO),
                   Product([f1, Symbol("Jinv_21", GEO)])])
    PS0 = Product([Symbol("Jinv_22", GEO),
                   Sum([P00, P01])])
    PS1 = Product([Product([f0, Symbol("Jinv_02", GEO)]),
                   Sum([P20, P21])])
    SP = Sum([PS0, PS1])

    PFx = PF.expand()
    E0x = E0.expand()
    E1x = E1.expand()
    E2x = E2.expand()
    E3x = E3.expand()
    E4x = E4.expand()
    P4x = P4.expand()
    P5x = P5.expand()
    P6x = P6.expand()
    S4x = S4.expand()
    SPx = SP.expand()

    Jinv_00, Jinv_01, Jinv_10, Jinv_02, Jinv_20, Jinv_22, Jinv_21, W1, det = [1, 2, 3, 4, 5, 6, 7, 8, 9]

    assert round(eval(str(SP)) - eval(str(SPx)), 10) == 0.0
    assert round(eval(str(E0)) - eval(str(E0x)), 10) == 0.0
    assert round(eval(str(E1)) - eval(str(E1x)), 10) == 0.0
    assert round(eval(str(E2)) - eval(str(E2x)), 10) == 0.0
    assert round(eval(str(E3)) - eval(str(E3x)), 10) == 0.0
    assert round(eval(str(E4)) - eval(str(E4x)), 10) == 0.0
    assert round(eval(str(SP)) - eval(str(SPx)), 10) == 0.0
    assert round(eval(str(P4)) - eval(str(P4x)), 10) == 0.0
    assert round(eval(str(P5)) - eval(str(P5x)), 10) == 0.0
    assert P6x == s1
    assert S4x == s1
    assert PF.ops() == 6
    assert PFx.ops() == 5
    assert E0.ops() == 4
    assert E0x.ops() == 6
    assert E1.ops() == 7
    assert E1x.ops() == 3
    assert E2.ops() == 1
    assert E2x.ops() == 0
    assert E3.ops() == 5
    assert E3x.ops() == 5
    assert E4.ops() == 10
    assert E4x.ops() == 6
    assert SP.ops() == 11
    assert SPx.ops() == 13
    assert P4.ops() == 2
    assert P4x.ops() == 3
    assert P5.ops() == 5
    assert P5x.ops() == 9



# -*- coding: utf-8 -*-
"""
Tests of generic code formatting utilities,
which focus on the overall structure of the code,
e.g. indentation, curly brace blocks, control flow
structures like loops, and class and function definitions.
Some of this is C++ specific, some is more generic.
"""

import pytest
from ffc.uflacs.language.cnodes import *


def test_format_basics():
    # Reproduce a string
    assert format_indented_lines("string") == "string"

    # Basic strings with indentation
    assert format_indented_lines("string", 1) == "    string"
    assert format_indented_lines("string", 2) == "        string"

    # Multiline strings with indentation
    assert format_indented_lines("fee\nfie\nfoe", level=1) == "    fee\n    fie\n    foe"

    # One item lists are the same as strings
    assert format_indented_lines(["hello"]) == "hello"
    assert format_indented_lines(["hello"], 1) == "    hello"
    assert format_indented_lines(["hello"], 2) == "        hello"

    # Lists are joined with newlines
    assert format_indented_lines(["hello", "world"]) == "hello\nworld"
    assert format_indented_lines(["hello", "world"], 1) == "    hello\n    world"
    assert format_indented_lines(["hello", "world"], 2) == "        hello\n        world"

    # Tuples are joined with newlines
    assert format_indented_lines(("hello", "world")) == "hello\nworld"
    assert format_indented_lines(("hello", "world"), 1) == "    hello\n    world"
    assert format_indented_lines(("hello", "world"), 2) == "        hello\n        world"
    code = format_indented_lines(("hello", "world"), 2)
    assert code == "        hello\n        world"


def test_format_indented_lines():
    # Strings and lists can be put in Indented containers
    assert format_indented_lines(Indented("hei")) == "    hei"
    code = format_indented_lines(Indented(["hei", Indented("verden")]))
    assert code == "    hei\n        verden"
    code = format_indented_lines(["{", Indented("fee\nfie\nfoe"), "}"])
    assert code == "{\n    fee\n    fie\n    foe\n}"


def xtest_format_blocks():
    # A Scope is an indented body with brackets before and after
    code = format_code(Scope("fee\nfie\nfoe"))
    assert code == "{\n    fee\n    fie\n    foe\n}"
    # A Namespace is a 'namespace foo' line before a block
    code = format_code(Namespace("bar", "fee\nfie\nfoe"))
    assert code == "namespace bar\n{\n    fee\n    fie\n    foe\n}"

    # Making a for loop
    code = format_code(["for (iq...)", Scope("foo(iq);")])
    assert code == "for (iq...)\n{\n    foo(iq);\n}"
    # Making a do loop
    code = format_code(["iq = 0;", "do", (Scope("foo(iq);"), " while (iq < nq);")])
    assert code == "iq = 0;\ndo\n{\n    foo(iq);\n} while (iq < nq);"


def xtest_format_class():
    # Making a class declaration
    assert format_code(Class('Car')) == 'class Car\n{\n};'
    code = format_code(Class('Car', superclass='Vehicle'))
    assert code == 'class Car: public Vehicle\n{\n};'
    code = format_code(Class('Car', public_body='void eval()\n{\n}'))
    assert code == 'class Car\n{\npublic:\n    void eval()\n    {\n    }\n};'
    code = format_code(Class('Car', protected_body='void eval()\n{\n}'))
    assert code == 'class Car\n{\nprotected:\n    void eval()\n    {\n    }\n};'
    code = format_code(Class('Car', private_body='void eval()\n{\n}'))
    assert code == 'class Car\n{\nprivate:\n    void eval()\n    {\n    }\n};'


def xtest_format_template_argument_list():
    def t(args, mlcode, slcode):
        code = format_code(TemplateArgumentList(args, False))
        assert code == slcode
        code = format_code(TemplateArgumentList(args, True))
        assert code == mlcode
    t(('A',), '<\n    A\n>', '<A>')
    t(('A', 'B'), '<\n    A,\n    B\n>', '<A, B>')


def xtest_format_templated_type():
    code = format_code(Type('Foo'))
    assert code == 'Foo'
    code = format_code(Type('Foo', ('int', '123')))
    assert code == 'Foo<int, 123>'
    code = format_code(Type('Foo', ('int', Type('Bar', ('123', 'float')))))
    assert code == 'Foo<int, Bar<123, float> >'


def xtest_format_typedef():
    assert format_code(TypeDef('int', 'myint')) == 'typedef int myint;'
    code = format_code(TypeDef(Type('Foo', ('int', Type('Bar', ('123', 'float')))), 'Thing'))
    assert code == 'typedef Foo<int, Bar<123, float> > Thing;'


def xtest_format_template_class():
    expected = "template<typename T, typename R>\nclass MyClass\n{\npublic:\n    void hello(int world) {}\n};"

    code = [('template', TemplateArgumentList(('typename T', 'typename R'), False)),
            Class('MyClass', public_body='void hello(int world) {}')]
    assert format_code(code) == expected

    code = Class('MyClass', public_body='void hello(int world) {}',
                 template_arguments=('typename T', 'typename R'))
    assert format_code(code) == expected


def test_format_variable_decl():
    code = VariableDecl("double", "foo")
    expected = "double foo;"
    assert str(code) == expected


def test_literal_cexpr_value_conversion():
    assert bool(LiteralBool(True)) is True
    assert bool(LiteralBool(False)) is False
    assert int(LiteralInt(2)) == 2
    assert float(LiteralFloat(2.0)) == 2.0
    # assert complex(LiteralFloat(2.0+4.0j)) == 2.0+4.0j


def test_format_array_decl():
    expected = "double foo[3];"
    code = ArrayDecl("double", "foo", 3)
    assert str(code) == expected
    code = ArrayDecl("double", "foo", (3,))
    assert str(code) == expected
    decl = code

    expected = "foo[1]"
    code = ArrayAccess(decl, 1)
    assert str(code) == expected
    code = ArrayAccess(decl, (1,))
    assert str(code) == expected

    expected = "foo[0]"
    code = ArrayAccess(decl, 0)
    assert str(code) == expected
    with pytest.raises(ValueError):
        ArrayAccess(decl, 3)

    code = ArrayDecl("double", "foo", (3, 4))
    expected = "double foo[3][4];"
    assert str(code) == expected
    decl = code

    expected = "foo[0][1]"
    code = ArrayAccess(decl, (0, 1))
    assert str(code) == expected
    expected = "foo[2][3]"
    code = ArrayAccess(decl, (2, 3))
    assert str(code) == expected

    with pytest.raises(ValueError):
        ArrayAccess(decl, 0)
    with pytest.raises(ValueError):
        ArrayAccess(decl, (0, 4))
    with pytest.raises(ValueError):
        ArrayAccess(decl, (3, 0))
    with pytest.raises(ValueError):
        ArrayAccess(decl, (3, -1))


def test_format_array_def():
    expected = "double foo[3] = { 1.0, 2.0, 3.0 };"
    code = ArrayDecl("double", "foo", 3, [1.0, 2.0, 3.0])
    assert str(code) == expected

    expected = """double foo[2][3] =
    { { 1.0, 2.0, 3.0 },
      { 6.0, 5.0, 4.0 } };"""
    code = ArrayDecl("double", "foo", (2, 3), [[1.0, 2.0, 3.0], [6.0, 5.0, 4.0]])
    assert str(code) == expected

    expected = """double foo[2][2][3] =
    { { { 1.0, 2.0, 3.0 },
        { 6.0, 5.0, 4.0 } },
      { { 1.0, 2.0, 3.0 },
        { 6.0, 5.0, 4.0 } } };"""
    code = ArrayDecl("double", "foo", (2, 2, 3), [[[1.0, 2.0, 3.0], [6.0, 5.0, 4.0]],
                                                  [[1.0, 2.0, 3.0], [6.0, 5.0, 4.0]]])
    assert str(code) == expected


def test_format_array_def_zero():
    expected = "double foo[3] = {};"
    code = ArrayDecl("double", "foo", 3, [0.0, 0.0, 0.0])
    assert str(code) == expected
    code = ArrayDecl("double", "foo", (3,), [0.0, 0.0, 0.0])
    assert str(code) == expected

    expected = "double foo[2][3] = {};"
    code = ArrayDecl("double", "foo", (2, 3), [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
    assert str(code) == expected


def xtest_class_with_arrays():
    adecl = ArrayDecl('double', 'phi', (3, 5))
    aacc = ArrayAccess(adecl, (2, 1))
    expected = "phi[2][1]"
    assert str(aacc) == expected

    decls = [VariableDecl('double', 'foo'),
             VariableDecl('int', 'bar'),
             adecl,
             ]
    classcode = Class('Car', public_body=decls)

    expected_decls = [str(decl) for decl in decls]
    expected = 'class Car\n{\npublic:\n    %s\n};' % '\n    '.join(expected_decls)
    actual = str(classcode)
    assert actual == expected


def test_class_array_access():
    vname = 'phi'
    shape = (3, 4)
    component = ('i0', 2)
    decl = ArrayDecl('double', vname, shape)
    dcode = str(decl)
    acode = str(ArrayAccess(decl, component))
    assert dcode == 'double phi[3][4];'
    assert acode == 'phi[i0][2]'


def test_while_loop():
    code = While("--k < 3", [])
    actual = str(code)
    expected = "while (--k < 3)\n{\n}"
    assert actual == expected

    code = While("--k < 3", body=["ting;", "tang;"])
    actual = str(code)
    expected = "while (--k < 3)\n{\n    ting;\n    tang;\n}"
    assert actual == expected


def test_for_loop():
    code = For("int i = 0", "i < 3", "++i", [])
    actual = str(code)
    expected = "for (int i = 0; i < 3; ++i)\n{\n}"
    assert actual == expected

    code = For("int i = 0", "i < 3", "++i", body=["ting;", "tang;"])
    actual = str(code)
    expected = "for (int i = 0; i < 3; ++i)\n{\n    ting;\n    tang;\n}"
    assert actual == expected


def test_for_range():
    code = ForRange("i", 0, 3, [])
    actual = str(code)
    expected = "for (int i = 0; i < 3; ++i)\n{\n}"
    assert actual == expected

    code = ForRange("i", 0, 3, body=["ting;", "tang;"])
    actual = str(code)
    expected = "for (int i = 0; i < 3; ++i)\n{\n    ting;\n    tang;\n}"
    assert actual == expected



# -*- coding: utf-8 -*-
"""
Tests of dolfin Expression formatting.
"""

import pytest

import ufl
from ufl.constantvalue import as_ufl

import numpy

import hashlib

from ffc.uflacs.params import default_parameters
from ffc.uflacs.backends.dolfin.expression import format_dolfin_expression
from ffc.uflacs.backends.dolfin.dolfin_compiler import compile_dolfin_expression_body


# FIXME: Make tests with dolfin optional
@pytest.fixture
def dolfin():
    import dolfin
    return dolfin


# For test comparisons
dolfin_expression_minimal = """\
class MyExpression: public Expression
{
public:

    virtual void eval(Array<double>& values,
                      const Array<double>& x) const
    {

    }
};"""

dolfin_expression_with_constructor1 = """\
class MyExpression: public Expression
{
public:
    MyExpression(): Expression(2) {}

    virtual void eval(Array<double>& values,
                      const Array<double>& x) const
    {

    }
};"""

dolfin_expression_with_constructor2 = """\
class MyExpression: public Expression
{
public:
    MyExpression(): Expression(2, 3) {}

    virtual void eval(Array<double>& values,
                      const Array<double>& x) const
    {

    }
};"""

dolfin_expression_with_members = """\
class MyExpression: public Expression
{
public:

    virtual void eval(Array<double>& values,
                      const Array<double>& x) const
    {

    }

    double c;
    boost::shared_ptr<MeshFunction<double> > mf;
    boost::shared_ptr<GenericFunction> gf;
    boost::shared_ptr<Function> f;
};"""

dolfin_expression_with_members2 = """\
class MyExpression: public Expression
{
public:

    virtual void eval(Array<double>& values,
                      const Array<double>& x) const
    {

    }

    double c0;
    double c1;
    boost::shared_ptr<MeshFunction<double> > mf0;
    boost::shared_ptr<MeshFunction<double> > mf1;
    boost::shared_ptr<GenericFunction> gf0;
    boost::shared_ptr<GenericFunction> gf1;
    boost::shared_ptr<Function> f0;
    boost::shared_ptr<Function> f1;
};"""

dolfin_expression_with_cell = """\
class MyExpression: public Expression
{
public:

    void eval(Array<double>& values, const Array<double>& x,
              const ufc::cell& ufc_cell) const
    {
        // Emulate tabulate_tensor arguments
        const ufc::cell& c = ufc_cell;
        Array<double> & A = values;

        //dolfin_assert(ufc_cell.local_facet >= 0);

        values[0] = 1.23;
    }
};"""


other_stuff = """
        // Getting cell data
        const uint gd = cell.geometric_dimension;
        const uint td = cell.topological_dimension;
        const uint cell_index = cell.entity_indices[td][0];

        // Tabulation of mesh function values
        double mf0v = (*mf0)[cell_index];

        // Evaluation of functions
        Array<double> w0v(1);
        w0->eval(w0v, x, cell);

        // Writing output values
        values[0] = w0v[0]*mf0v;
        values[1] = w0v[0]+mf0v;
"""

#class DolfinExpressionFormatterTest(UflTestCase):
#"""Tests of Dolfin C++ Expression formatting utilities."""

def test_dolfin_expression_with_cell(dolfin):
    e = dolfin.Expression(cppcode=dolfin_expression_with_cell)
    assert e.ufl_shape == ()
    values = numpy.zeros((1,))
    x = numpy.asarray((0.2, 0.2))
    mesh = dolfin.UnitSquareMesh(1, 1)
    cell = dolfin.Cell(mesh, 0)
    e.eval_cell(values, x, cell)
    assert values[0] == 1.23
    #assert e((1.0,2.0)) == 1.23 # FAILS! Need dolfin to dispatch this to eval_cell with the proper Cell found from x for this to work.

def test_dolfin_expression_constructors():
    code = format_dolfin_expression(classname="MyExpression",
                                    shape=())
    assert code == dolfin_expression_minimal
    code = format_dolfin_expression(classname="MyExpression",
                                    shape=(2,))
    assert code == dolfin_expression_with_constructor1
    code = format_dolfin_expression(classname="MyExpression",
                                    shape=(2, 3))
    assert code == dolfin_expression_with_constructor2

def test_dolfin_expression_members():
    code = format_dolfin_expression(classname="MyExpression",
                                    shape=(),
                                    constants=('c',),
                                    mesh_functions=('mf',),
                                    generic_functions=('gf',),
                                    functions=('f',),
                                    )
    assert code == dolfin_expression_with_members
    code = format_dolfin_expression(classname="MyExpression",
                                    shape=(),
                                    constants=('c0', 'c1'),
                                    mesh_functions=('mf0', 'mf1'),
                                    generic_functions=('gf0', 'gf1'),
                                    functions=('f0', 'f1'),
                                    )
    assert code == dolfin_expression_with_members2

def test_explicit_dolfin_expression_compilation(dolfin):
    code = format_dolfin_expression(classname="MyExpression",
                                    shape=(2,),
                                    eval_body=['values[0] = x[0];',
                                               'values[1] = x[1];'],
                                    constants=('c0', 'c1'),
                                    mesh_functions=('mf0', 'mf1'),
                                    generic_functions=('gf0', 'gf1'),
                                    functions=('f0', 'f1'),
                                    )
    expr = dolfin.Expression(cppcode=code)
    assert hasattr(expr, 'c0')
    assert hasattr(expr, 'c1')
    assert hasattr(expr, 'mf0')
    assert hasattr(expr, 'mf1')
    assert hasattr(expr, 'gf0')
    assert hasattr(expr, 'gf1')
    assert hasattr(expr, 'f0')
    assert hasattr(expr, 'f1')
    #dolfin.plot(expr, mesh=dolfin.UnitSquareMesh(10,10), interactive=True)


def flattened_nonempty_lines(lines):
    res = []
    for l in lines:
        if isinstance(l, list):
            res.extend(flattened_nonempty_lines(l))
        elif l:
            res.append(l)
        else:
            pass
    return res

#class Ufl2DolfinExpressionCompilerTest(UflTestCase):
#"""Tests of Dolfin C++ Expression compilation from UFL expressions."""

def check_dolfin_expression_compilation(uexpr, expected_lines, expected_values, members={}):
    parameters = default_parameters()
    # Compile expression
    compiled_lines, member_names = compile_dolfin_expression_body(uexpr, parameters)

    # Check expected compilation output
    if flattened_nonempty_lines(compiled_lines) != expected_lines:
        print('\n'*5)
        print("Upcoming failure, expected lines:")
        print('\n'.join(expected_lines))
        print("\nActual lines:")
        print('\n'.join(flattened_nonempty_lines(compiled_lines)))
        print('\n'*5)
    assert flattened_nonempty_lines(compiled_lines) == expected_lines

    # Add debugging lines to print scalar expressions at end of Expression eval call
    if 0:
        printme = ["x[0]", "values[0]"]
        compiled_lines += ['std::cout '] + ['       << "%s = " << %s << std::endl' % (pm, pm) for pm in printme] + ['<< std::endl;']

    # Wrap it in a dolfin::Expression class
    shape = uexpr.ufl_shape
    name = "MyExpression_%s" % hashlib.md5(str(hash(uexpr))).hexdigest()
    code = format_dolfin_expression(classname=name,
                                    shape=shape,
                                    eval_body=compiled_lines,
                                    **member_names)
    if 0:
        print("SKIPPING REST OF TEST")
        return

    # Try to compile it with DOLFIN
    import dolfin
    dexpr = dolfin.Expression(cppcode=code)

    # Connect compiled dolfin::Expression object with dolfin::Function instances
    for name, value in members.items():
        setattr(dexpr, name, value)

    # Evaluate and assert compiled value!
    for x, v in expected_values:
        u = dexpr(x) # TODO: Not sure if this works
        if not hasattr(u, '__len__'):
            u = (u,)
        diff = numpy.array(u) - numpy.array(v)
        if numpy.dot(diff, diff) > 1e-13:
            print("u =", u)
            print("v =", v)
            print("diff =", diff)
        assert numpy.dot(diff, diff) < 1e-14

def test_dolfin_expression_compilation_of_scalar_literal():
    # Define some literal ufl expression
    uexpr = as_ufl(3.14)

    # Define expected output from compilation
    expected_lines = ['double s[1];',
                      's[0] = 3.14;',
                      'values[0] = s[0];']

    # Define expected evaluation values: [(x,value), (x,value), ...]
    expected_values = [((0.0, 0.0), (3.14,)),
                       ((0.6, 0.7), (3.14,)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values)

def test_dolfin_expression_compilation_of_vector_literal():
    # Define some literal ufl expression
    uexpr = ufl.as_vector((1.23, 7.89))

    # Define expected output from compilation
    expected_lines = ['double s[2];',
                      's[0] = 1.23;',
                      's[1] = 7.89;',
                      'values[0] = s[0];',
                      'values[1] = s[1];']

    # Define expected evaluation values: [(x,value), (x,value), ...]
    expected_values = [((0.0, 0.0), (1.23, 7.89)),
                       ((0.6, 0.7), (1.23, 7.89)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values)

def test_dolfin_expression_compilation_of_x():
    # Define some ufl expression
    x = ufl.SpatialCoordinate(ufl.triangle)
    uexpr = 2*x

    # Define expected output from compilation
    expected_lines = ['double s[2];',
                      's[0] = 2 * x[0];',
                      's[1] = 2 * x[1];',
                      'values[0] = s[0];',
                      'values[1] = s[1];']

    # Define expected evaluation values: [(x,value), (x,value), ...]
    expected_values = [((0.0, 0.0), (0.0, 0.0)),
                       ((0.6, 0.7), (1.2, 1.4)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values)

def test_dolfin_expression_compilation_of_coefficient(dolfin):

    # Define some ufl expression with PyDOLFIN coefficients
    mesh = dolfin.UnitSquareMesh(3, 3)
    # Using quadratic element deliberately for accuracy
    V = dolfin.FunctionSpace(mesh, "CG", 2)
    u = dolfin.Function(V)
    u.interpolate(dolfin.Expression("x[0]*x[1]"))
    uexpr = 2*u

    # Define expected output from compilation
    expected_lines = ['double s[1];',
                      'Array<double> v_w0(1);',
                      'w0->eval(v_w0, x);',
                      's[0] = 2 * v_w0[0];',
                      'values[0] = s[0];']

    # Define expected evaluation values: [(x,value), (x,value), ...]
    expected_values = [((0.0, 0.0), (0.0,)),
                       ((0.6, 0.7), (2*0.6*0.7,)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values, members={'w0':u})

def test_dolfin_expression_compilation_of_algebraic_operators(dolfin):

    # Define some PyDOLFIN coefficients
    mesh = dolfin.UnitSquareMesh(3, 3)
    # Using quadratic element deliberately for accuracy
    V = dolfin.FunctionSpace(mesh, "CG", 2)
    u = dolfin.Function(V)
    u.interpolate(dolfin.Expression("x[0]*x[1]"))

    # Define ufl expression with algebraic operators
    uexpr = (2*u + 3.14*u**2) / 4

    # Define expected output from compilation
    # TODO: Test with and without variables
    # Fully split version:
    expected_lines = ['double s[4];',
                      'Array<double> v_w0(1);',
                      'w0->eval(v_w0, x);',
                      's[0] = 2 * v_w0[0];',
                      's[1] = pow(v_w0[0], 2);',
                      's[2] = 3.14 * s[1];',
                      's[3] = s[2] / 4;',
                      'values[0] = s[3];']
    # Oneliner version (no reuse in this expression):
    expected_lines = ['double s[1];',
                      'Array<double> v_w0(1);',
                      'w0->eval(v_w0, x);',
                      's[0] = (3.14 * pow(v_w0[0], 2) + 2 * v_w0[0]) / 4;',
                      'values[0] = s[0];']

    # Define expected evaluation values: [(x,value), (x,value), ...]
    x, y = 0.6, 0.7; expected = (2*x*y + 3.14*(x*y)**2) / 4
    expected_values = [((0.0, 0.0), (0.0,)),
                       ((x, y), (expected,)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values, members={'w0':u})


def test_dolfin_expression_compilation_of_math_functions(dolfin):

    # Define some PyDOLFIN coefficients
    mesh = dolfin.UnitSquareMesh(3, 3)
    # Using quadratic element deliberately for accuracy
    V = dolfin.FunctionSpace(mesh, "CG", 2)
    u = dolfin.Function(V)
    u.interpolate(dolfin.Expression("x[0]*x[1]"))
    w0 = u

    # Define ufl expression with math functions
    v = abs(ufl.cos(u))/2 + 0.02
    uexpr = ufl.sin(u) + ufl.tan(v) + ufl.exp(u) + ufl.ln(v) + ufl.atan(v) + ufl.acos(v) + ufl.asin(v)

    #print dolfin.assemble(uexpr**2*dolfin.dx, mesh=mesh) # 11.7846508409

    # Define expected output from compilation
    ucode = 'v_w0[0]'
    vcode = '0.02 + fabs(cos(v_w0[0])) / 2'
    funcs = 'asin(%(v)s) + (acos(%(v)s) + (atan(%(v)s) + (log(%(v)s) + (exp(%(u)s) + (sin(%(u)s) + tan(%(v)s))))))'
    oneliner = funcs % {'u':ucode, 'v':vcode}

    # Oneliner version (ignoring reuse):
    expected_lines = ['double s[1];',
                      'Array<double> v_w0(1);',
                      'w0->eval(v_w0, x);',
                      's[0] = %s;' % oneliner,
                      'values[0] = s[0];']
    #cppcode = format_dolfin_expression(classname="DebugExpression", shape=(), eval_body=expected_lines)
    #print '-'*100
    #print cppcode
    #print '-'*100
    #dolfin.plot(dolfin.Expression(cppcode=cppcode, mesh=mesh))
    #dolfin.interactive()

    # Split version (handles reuse of v, no other reuse):
    expected_lines = ['double s[2];',
                      'Array<double> v_w0(1);',
                      'w0->eval(v_w0, x);',
                      's[0] = %s;' % (vcode,),
                      's[1] = %s;' % (funcs % {'u':ucode,'v':'s[0]'},),
                      'values[0] = s[1];']

    # Define expected evaluation values: [(x,value), (x,value), ...]
    import math
    x, y = 0.6, 0.7
    u = x*y
    v = abs(math.cos(u))/2 + 0.02
    v0 = .52
    expected0 = math.tan(v0) + 1 + math.log(v0) + math.atan(v0) + math.acos(v0) + math.asin(v0)
    expected = math.sin(u) + math.tan(v) + math.exp(u) + math.log(v) + math.atan(v) + math.acos(v) + math.asin(v)
    expected_values = [((0.0, 0.0), (expected0,)),
                       ((x, y), (expected,)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values, members={'w0':w0})

def test_dolfin_expression_compilation_of_dot_with_index_notation(dolfin):

    # Define some PyDOLFIN coefficients
    mesh = dolfin.UnitSquareMesh(3, 3)
    # Using quadratic element deliberately for accuracy
    V = dolfin.VectorFunctionSpace(mesh, "CG", 2)
    u = dolfin.Function(V)
    u.interpolate(dolfin.Expression(("x[0]", "x[1]")))

    # Define ufl expression with the simplest possible index notation
    uexpr = ufl.dot(u, u) # Needs expand_compounds, not testing here

    uexpr = u[0]*u[0] + u[1]*u[1] # Works

    i, = ufl.indices(1)
    uexpr = u[i]*u[i] # Works

    # Define expected output from compilation
    # Oneliner version (no reuse in this expression):
    xc, yc = ('v_w0[0]', 'v_w0[1]')
    expected_lines = ['double s[1];',
                      'Array<double> v_w0(2);',
                      'w0->eval(v_w0, x);',
                      's[0] = pow(%(x)s, 2) + pow(%(y)s, 2);' % {'x':xc, 'y':yc},
                      'values[0] = s[0];']

    # Define expected evaluation values: [(x,value), (x,value), ...]
    x, y = (0.6, 0.7)
    expected = (x*x+y*y)
    expected_values = [((0.0, 0.0), (0.0,)),
                       ((x, y), (expected,)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values, members={'w0':u})

def test_dolfin_expression_compilation_of_index_notation(dolfin):

    # Define some PyDOLFIN coefficients
    mesh = dolfin.UnitSquareMesh(3, 3)
    # Using quadratic element deliberately for accuracy
    V = dolfin.VectorFunctionSpace(mesh, "CG", 2)
    u = dolfin.Function(V)
    u.interpolate(dolfin.Expression(("x[0]", "x[1]")))

    # Define ufl expression with index notation
    i, j, k, l = ufl.indices(4)
    uexpr = ((u[0]*u + u[1]*u)[i]*u[j])*(u[i]*u[j])

    # Define expected output from compilation
    # Split version (reusing all product terms correctly!):
    xc, yc = ('v_w0[0]', 'v_w0[1]')
    svars = {'a':'s[2]', 'b':'s[4]', 'xx':'s[0]', 'yy':'s[3]', 'xy':'s[1]', 'x':xc, 'y':yc}
    expected_lines = ['double s[6];',
                      'Array<double> v_w0(2);',
                      'w0->eval(v_w0, x);',
                      's[0] = pow(%(x)s, 2);' % svars,
                      's[1] = %(x)s * %(y)s;' % svars,
                      's[2] = %(xx)s + %(xy)s;' % svars,
                      's[3] = pow(%(y)s, 2);' % svars,
                      's[4] = %(yy)s + %(xy)s;' % svars,
                      's[5] = (%(xx)s * (%(x)s * %(a)s) + %(xy)s * (%(x)s * %(b)s)) + (%(yy)s * (%(y)s * %(b)s) + %(xy)s * (%(y)s * %(a)s));' % svars,
                      'values[0] = s[5];']
    # Making the above line correct was a nightmare because of operator sorting in ufl... Hoping it stays stable!

    # Define expected evaluation values: [(x,value), (x,value), ...]
    x, y = (0.6, 0.7)
    a = x*x+y*x
    b = x*y+y*y
    expected = (a*x)*(x*x) + (a*y)*(x*y) + (b*x)*(y*x) + (b*y)*(y*y)
    expected_values = [((0.0, 0.0), (0.0,)),
                       ((x, y), (expected,)),
                       ]

    # Execute all tests
    check_dolfin_expression_compilation(uexpr, expected_lines, expected_values, members={'w0':u})



# -*- coding: utf-8 -*-
# Copyright (C) 2009-2016 Anders Logg and Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

# Note: Most of the code in this file is a direct translation from the old implementation in FFC


from ffc.uflacs.backends.ufc.generator import ufc_generator
from ffc.uflacs.backends.ufc.utils import generate_return_new_switch


class ufc_dofmap(ufc_generator):
    def __init__(self):
        ufc_generator.__init__(self, "dofmap")

    def topological_dimension(self, L, ir):
        "Default implementation of returning topological dimension fetched from ir."
        tdim = ir["topological_dimension"]
        return L.Return(L.LiteralInt(tdim))

    def needs_mesh_entities(self, L, ir):
        d = L.Symbol("d")
        nme = ir["needs_mesh_entities"]
        cases = [(L.LiteralInt(dim), L.Return(L.LiteralBool(need)))
                 for dim, need in enumerate(nme)]
        default = L.Return(L.LiteralBool(False))
        return L.Switch(d, cases, default=default, autoscope=False, autobreak=False)

    def global_dimension(self, L, ir):
        # A list of num dofs per entity
        num_entity_dofs, num_global_dofs = ir["global_dimension"]

        # Input array
        entities = L.Symbol("num_global_entities")

        # Accumulate number of dofs per entity times given number of entities in mesh
        dimension = sum(L.LiteralInt(num)*entities[dim] for dim, num in enumerate(num_entity_dofs) if num > 0)

        # Add global dofs if any
        if num_global_dofs:
            dimension = dimension + L.LiteralInt(num_global_dofs)

        return L.Return(dimension)

    def num_element_dofs(self, L, ir):
        value = ir["num_element_dofs"]
        return L.Return(L.LiteralInt(value))

    def num_facet_dofs(self, L, ir):
        value = ir["num_facet_dofs"]
        return L.Return(L.LiteralInt(value))

    def num_entity_dofs(self, L, ir):
        d = L.Symbol("d")
        values = ir["num_entity_dofs"]
        cases = [(i, L.Return(L.LiteralInt(value))) for i, value in enumerate(values)]
        default = L.Return(L.LiteralInt(0))
        return L.Switch(d, cases, default=default)

    def num_entity_closure_dofs(self, L, ir):
        d = L.Symbol("d")
        values = ir["num_entity_closure_dofs"]
        cases = [(i, L.Return(L.LiteralInt(value))) for i, value in enumerate(values)]
        default = L.Return(L.LiteralInt(0))
        return L.Switch(d, cases, default=default)

    def tabulate_dofs(self, L, ir):

        # Input arguments
        entity_indices = L.Symbol("entity_indices")
        num_mesh_entities = L.Symbol("num_global_entities")

        # Output arguments
        dofs_variable = L.Symbol("dofs")


        ir = ir["tabulate_dofs"]
        if ir is None: # What is this supposed to mean?
            return L.Assign(dofs_variable[0], 0)

        # Extract representation
        #(dofs_per_element, num_dofs_per_element, need_offset, fakes) = ir # names from original ffc code
        (subelement_dofs, num_dofs_per_subelement, need_offset, is_subelement_real) = ir

        #len(is_subelement_real) == len(subelement_dofs) == len(num_dofs_per_subelement) == number of combined (flattened) subelements?
        #is_subelement_real[subelement_number] is bool, is subelement Real?
        #num_dofs_per_subelement[subelement_number] is int, number of dofs for each (flattened) subelement
        #subelement_dofs[subelement_number] is list of entity_dofs for each (flattened) subelement
        #len(entity_dofs) == tdim+1
        #len(entity_dofs[cell_entity_dim]) == "num_cell_entities[dim]" (i.e. 3 vertices, 3 edges, 1 face for triangle)
        #len(entity_dofs[cell_entity_dim][entity_index[dim][k]]) == number of dofs for this entity
        #num = entity_dofs[dim]

        #entity_dofs =? entity_dofs[d][i][:] # dofs on entity (d,i)

        # Collect code pieces in list
        code = []

        # Declare offset if needed
        if need_offset:
            offset = L.Symbol("offset")
            code.append(L.VariableDecl("std::size_t", offset, value=0))
        else:
            offset = 0

        # Generate code for each element
        subelement_offset = 0
        for (subelement_index, entity_dofs) in enumerate(subelement_dofs):

            # Handle is_subelement_real (Space of reals)
            if is_subelement_real[subelement_index]:
                assert num_dofs_per_subelement[subelement_index] == 1
                code.append(L.Assign(dofs_variable[subelement_offset], offset))
                if need_offset:
                    code.append(L.AssignAdd(offset, 1))
                subelement_offset += 1
                continue

            # Generate code for each degree of freedom for each dimension
            for (cell_entity_dim, dofs_on_cell_entity) in enumerate(entity_dofs):
                num_dofs_per_mesh_entity = len(dofs_on_cell_entity[0])
                assert all(num_dofs_per_mesh_entity == len(dofs)
                           for dofs in dofs_on_cell_entity)

                # Ignore if no dofs for this dimension
                if num_dofs_per_mesh_entity == 0:
                    continue

                # For each cell entity of this dimension
                for (cell_entity_index, dofs) in enumerate(dofs_on_cell_entity):
                    # dofs is a list of the local dofs that live on this cell entity

                    # find offset for this particular mesh entity
                    entity_offset = len(dofs) * entity_indices[cell_entity_dim, cell_entity_index]

                    for (j, dof) in enumerate(dofs):
                        # dof is the local dof index on the subelement
                        # j is the local index of dof among the dofs on this particular cell/mesh entity
                        local_dof_index = subelement_offset + dof
                        global_dof_index = offset + entity_offset + j
                        code.append(L.Assign(dofs_variable[local_dof_index], global_dof_index))

                # Update offset corresponding to mesh entity:
                if need_offset:
                    code.append(L.AssignAdd(offset, num_dofs_per_mesh_entity * num_mesh_entities[cell_entity_dim]))

            subelement_offset += num_dofs_per_subelement[subelement_index]

        return L.StatementList(code)

    def tabulate_facet_dofs(self, L, ir):
        all_facet_dofs = ir["tabulate_facet_dofs"]

        # Input arguments
        facet = L.Symbol("facet")
        dofs = L.Symbol("dofs")

        # For each facet, copy all_facet_dofs[facet][:] into output argument array dofs[:]
        cases = []
        for facet, single_facet_dofs in enumerate(all_facet_dofs):
            assigns = L.StatementList([L.Assign(dofs[i], dof) for (i, dof) in enumerate(single_facet_dofs)])
            cases.append((facet, assigns))

        return L.Switch(facet, cases, autoscope=False)

    def tabulate_entity_dofs(self, L, ir):
        entity_dofs, num_dofs_per_entity = ir["tabulate_entity_dofs"]

        # Output argument array
        dofs = L.Symbol("dofs")

        # Input arguments
        d = L.Symbol("d")
        i = L.Symbol("i")

        # TODO: Removed check for (d <= tdim + 1)
        tdim = len(num_dofs_per_entity) - 1

        # Generate cases for each dimension:
        all_cases = []
        for dim in range(tdim + 1):

            # Ignore if no entities for this dimension
            if num_dofs_per_entity[dim] == 0:
                continue

            # Generate cases for each mesh entity
            cases = []
            for entity in range(len(entity_dofs[dim])):
                casebody = []
                for (j, dof) in enumerate(entity_dofs[dim][entity]):
                    casebody += [L.Assign(dofs[j], dof)]
                cases.append((entity, L.StatementList(casebody)))

            # Generate inner switch
            # TODO: Removed check for (i <= num_entities-1)
            inner_switch = L.Switch(i, cases, autoscope=False)
            all_cases.append((dim, inner_switch))

        return L.Switch(d, all_cases, autoscope=False)

    def tabulate_entity_closure_dofs(self, L, ir):
        # Extract variables from ir
        entity_closure_dofs, entity_dofs, num_dofs_per_entity = \
            ir["tabulate_entity_closure_dofs"]

        # Output argument array
        dofs = L.Symbol("dofs")

        # Input arguments
        d = L.Symbol("d")
        i = L.Symbol("i")

        # TODO: Removed check for (d <= tdim + 1)
        tdim = len(num_dofs_per_entity) - 1

        # Generate cases for each dimension:
        all_cases = []
        for dim in range(tdim + 1):
            num_entities = len(entity_dofs[dim])

            # Generate cases for each mesh entity
            cases = []
            for entity in range(num_entities):
                casebody = []
                for (j, dof) in enumerate(entity_closure_dofs[(dim, entity)]):
                    casebody += [L.Assign(dofs[j], dof)]
                cases.append((entity, L.StatementList(casebody)))

            # Generate inner switch
            # TODO: Removed check for (i <= num_entities-1)
            inner_switch = L.Switch(i, cases, autoscope=False)
            all_cases.append((dim, inner_switch))

        return L.Switch(d, all_cases, autoscope=False)

    def num_sub_dofmaps(self, L, ir):
        value = ir["num_sub_dofmaps"]
        return L.Return(L.LiteralInt(value))

    def create_sub_dofmap(self, L, ir):
        i = L.Symbol("i")
        classnames = ir["create_sub_dofmap"]
        return generate_return_new_switch(L, i, classnames, factory=ir["jit"])



#!/usr/bin/env py.test
# -*- coding: utf-8 -*-

class MockBasicElementCodeGenerator:
    pass

def test_mock_element_generator(gtest):
    """Test that the basic element mock class works as assumed."""

    pre = """
    double x[3] = { 0.0, 0.1, 0.2 };
    """

    post = """
    ASSERT_EQ(xi[0], 0.0);
    ASSERT_EQ(xi[1], 0.1);
    ASSERT_EQ(xi[2], 0.2);
    """

    # FIXME: Define interface for mock element code generator using TDD

    code = """
    double xi[3];
    xi[0] = x[0];
    xi[1] = x[1];
    xi[2] = x[2];
    """

    gtest.add(pre + code + post)




# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testDGElastoDyn():
    expr = Product([
        Sum([
            Symbol("F0", IP),
            Symbol("F1", IP)
        ]),
        Fraction(
            Symbol("w4", GEO),
            Symbol("w3", GEO)
        ),
        Fraction(
            Product([
                Symbol("w2", GEO),
                Symbol("w5", GEO)
            ]),
            Symbol("w6", GEO)
        )
    ])

    expr_exp = expr.expand()

    expr_red = expr_exp.reduce_ops()

    F0, F1, w2, w3, w4, w5, w6 = (3.12, -8.1, -45.3, 17.5, 2.2, 5.3, 9.145)
    assert round(eval(str(expr)) - eval(str(expr_exp)), 10) == 0
    assert round(eval(str(expr)) - eval(str(expr_red)), 10) == 0
    assert expr.ops() == 6
    assert expr_exp.ops() == 11
    assert expr_red.ops() == 6



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

#import inspect
import re
#from string import Formatter

#from ufl import product
from ffc.log import error, warning
#from ffc.backends.ufc import *

import ffc.backends.ufc

from ffc.uflacs.language.format_lines import format_indented_lines
from ffc.uflacs.backends.ufc.templates import *

#__all__ = (["ufc_form", "ufc_dofmap", "ufc_finite_element", "ufc_integral"]
#           + ["ufc_%s_integral" % integral_type for integral_type in integral_types])


# These are all the integral types directly supported in ufc
from ffc.representation import ufc_integral_types


# These are the method names in ufc::form that are specialized for each integral type
integral_name_templates = (
    "max_%s_subdomain_id",
    "has_%s_integrals",
    "create_%s_integral",
    "create_default_%s_integral",
    )


class ufc_generator(object):
    """Common functionality for code generators producing ufc classes.

    The generate function is the driver for generating code for a class.
    It automatically extracts template keywords and inserts the results
    from calls to self.<keyword>(language, ir), or the value of ir[keyword]
    if there is no self.<keyword>.
    """
    def __init__(self, basename):
        ufc_templates = ffc.backends.ufc.templates
        self._header_template = ufc_templates[basename + "_header"]
        self._implementation_template = ufc_templates[basename + "_implementation"]
        self._combined_template = ufc_templates[basename + "_combined"]

        r = re.compile(r"%\(([a-zA-Z0-9_]*)\)")
        self._header_keywords = set(r.findall(self._header_template))
        self._implementation_keywords = set(r.findall(self._implementation_template))
        self._combined_keywords = set(r.findall(self._combined_template))

        self._keywords = sorted(self._header_keywords | self._implementation_keywords)

        # Do some ufc interface template checking, to catch bugs early when we change the ufc interface templates
        if set(self._keywords) != set(self._combined_keywords):
            a = set(self._header_keywords) - set(self._combined_keywords)
            b = set(self._implementation_keywords) - set(self._combined_keywords)
            c = set(self._combined_keywords) - set(self._keywords)
            msg = "Templates do not have matching sets of keywords:"
            if a:
                msg += "\n  Header template keywords '%s' are not in the combined template." % (sorted(a),)
            if b:
                msg += "\n  Implementation template keywords '%s' are not in the combined template." % (sorted(b),)
            if c:
                msg += "\n  Combined template keywords '%s' are not in the header or implementation templates." % (sorted(c),)
            error(msg)

    def generate_snippets(self, L, ir):
        "Generate code snippets for each keyword found in templates."
        snippets = {}
        for kw in self._keywords:
            # Check that attribute self.<keyword> is available
            if not hasattr(self, kw):
                error("Missing handler for keyword '%s' in class %s." % (kw, self.__class__.__name__))

            # Call self.<keyword>(L, ir) to get value
            method = getattr(self, kw)
            value = method(L, ir)

            # Indent body and format to str
            if isinstance(value, L.CStatement):
                value = L.Indented(value.cs_format())
                value = format_indented_lines(value)

            # Store formatted code in snippets dict
            snippets[kw] = value

        # Error checking (can detect some bugs early when changing the ufc interface)
        # Get all attributes of subclass class (skip "_foo")
        attrs = set(name for name in dir(self) if not (name.startswith("_") or name.startswith("generate")))
        # Get all attributes of this base class (skip "_foo" and "generate*")
        base_attrs = set(name for name in dir(ufc_generator) if not (name.startswith("_") or name.startswith("generate")))
        # The template keywords should not contain any names not among the class attributes
        missing = set(self._keywords) - attrs
        if missing:
            warning("*** Missing generator functions:\n%s" % ('\n'.join(map(str, sorted(missing))),))
        # The class attributes should not contain any names not among the template keywords
        # (this is strict, a useful check when changing ufc, but can be dropped)
        unused = attrs - set(self._keywords) - base_attrs
        if unused:
            warning("*** Unused generator functions:\n%s" % ('\n'.join(map(str, sorted(unused))),))

        # Return snippets, a dict of code strings
        return snippets

    def generate(self, L, ir, snippets=None):
        "Return composition of templates with generated snippets."
        if snippets is None:
            snippets = self.generate_snippets(L, ir)
        h = self._header_template % snippets
        cpp = self._implementation_template % snippets
        return h, cpp

    def classname(self, L, ir):
        "Return classname."
        return ir["classname"]

    def members(self, L, ir):
        "Return empty string. Override in classes that need members."
        if ir.get("members"):
            error("Missing generator function.")
        return ""

    def constructor(self, L, ir):
        "Return empty string. Override in classes that need constructor."
        if ir.get("constructor"):
            error("Missing generator function.")
        return ""

    def constructor_arguments(self, L, ir):
        "Return empty string. Override in classes that need constructor."
        if ir.get("constructor_arguments"):
            error("Missing generator function.")
        return ""

    def initializer_list(self, L, ir):
        "Return empty string. Override in classes that need constructor."
        if ir.get("initializer_list"):
            error("Missing generator function.")
        return ""

    def destructor(self, L, ir):
        "Return empty string. Override in classes that need destructor."
        if ir.get("destructor"):
            error("Missing generator function.")
        return ""

    def signature(self, L, ir):
        "Default implementation of returning signature string fetched from ir."
        sig = ir["signature"]
        return L.Return(L.LiteralString(sig))

    def create(self, L, ir):
        "Default implementation of creating a new object of the same type."
        classname = ir["classname"]
        return L.Return(L.New(classname))



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def test_Elasticity_Term():
    # expr:  0.25*W1*det*(FE0_C2_D001[0][j]*FE0_C2_D001[0][k]*Jinv_00*Jinv_21 + FE0_C2_D001[0][j]*FE0_C2_D001[0][k]*Jinv_00*Jinv_21)
    expr = Product([
        FloatValue(0.25), Symbol('W1', GEO), Symbol('det', GEO),
        Sum([Product([Symbol('FE0_C2_D001_0_j', BASIS), Symbol('FE0_C2_D001_0_k', BASIS),
                      Symbol('Jinv_00', GEO), Symbol('Jinv_21', GEO)]),
             Product([Symbol('FE0_C2_D001_0_j', BASIS), Symbol('FE0_C2_D001_0_k', BASIS),
                      Symbol('Jinv_00', GEO), Symbol('Jinv_21', GEO)])
             ])
    ])

    expr_exp = expr.expand()
    expr_red = expr_exp.reduce_ops()

    det, W1, Jinv_00, Jinv_21, FE0_C2_D001_0_j, FE0_C2_D001_0_k = [0.123 + i for i in range(6)]
    assert round(eval(str(expr)) - eval(str(expr_exp)), 10) == 0.0
    assert round(eval(str(expr)) - eval(str(expr_red)), 10) == 0.0
    assert expr.ops() == 10
    assert expr_exp.ops() == 6
    assert expr_red.ops() == 6

    # Generate code
    ip_consts = {}
    geo_consts = {}
    trans_set = set()

    start = time.time()
    opt_code = optimise_code(expr, ip_consts, geo_consts, trans_set)

    G = [eval(str(list(geo_consts.items())[0][0]))]
    assert round(eval(str(expr)) - eval(str(opt_code))) == 0.0



# -*- coding: utf-8 -*-
"""
Tests of table manipulation utilities.
"""

from __future__ import print_function

from ufl import triangle
from six import itervalues, iteritems
from six.moves import xrange as range
from ffc.uflacs.elementtables import equal_tables, strip_table_zeros, build_unique_tables, get_ffc_table_values

import numpy as np
default_tolerance = 1e-14


def old_strip_table_zeros(table, eps):
    "Quick fix translating new version to old behaviour to keep tests without modification."
    dofrange, dofmap, stripped_table = strip_table_zeros(table, False, eps)
    begin, end = dofrange
    return begin, end, stripped_table


def test_equal_tables():
    a = np.zeros((2, 3))
    b = np.zeros((2, 3))
    assert equal_tables(a, b, default_tolerance)

    a = np.ones((2, 3))
    b = np.zeros((2, 3))
    assert not equal_tables(a, b, default_tolerance)

    a = np.ones((2, 3))
    b = np.ones((3, 2))
    assert not equal_tables(a, b, default_tolerance)

    a = np.ones((2, 3))*1.1
    b = np.ones((2, 3))
    assert not equal_tables(a, b, default_tolerance)

    # Checking for all ones within given tolerance:
    eps = 1e-4
    a = np.ones((2, 3))*(1.0+eps)
    assert equal_tables(a, np.ones(a.shape), 10*eps)


def test_strip_table_zeros():
    # Can strip entire table:
    a = np.zeros((2, 3))
    e = np.zeros((2, 0))
    begin, end, b = old_strip_table_zeros(a, default_tolerance)
    assert begin == end # This is a way to check for all-zero table
    assert begin == 0
    assert end == 0
    assert equal_tables(b, e, default_tolerance)

    # Can keep entire nonzero table:
    a = np.ones((2, 3))
    e = np.ones((2, 3))
    begin, end, b = old_strip_table_zeros(a, default_tolerance)
    assert begin != end
    assert begin == 0
    assert end == a.shape[-1]
    assert equal_tables(b, e, default_tolerance)

    # Can strip one left side column:
    a = np.ones((2, 3))
    a[:, 0] = 0.0
    e = np.ones((2, 2))
    begin, end, b = old_strip_table_zeros(a, default_tolerance)
    assert begin != end
    assert begin == 1
    assert end == a.shape[-1]
    assert equal_tables(b, e, default_tolerance)

    # Can strip one right side column:
    a = np.ones((2, 3))
    a[:, 2] = 0.0
    e = np.ones((2, 2))
    begin, end, b = old_strip_table_zeros(a, default_tolerance)
    assert begin != end
    assert begin == 0
    assert end == a.shape[-1]-1
    assert equal_tables(b, e, default_tolerance)

    # Can strip two columns on each side:
    a = np.ones((2, 5))
    a[:, 0] = 0.0
    a[:, 1] = 0.0
    a[:, 3] = 0.0
    a[:, 4] = 0.0
    e = np.ones((2, 1))
    begin, end, b = old_strip_table_zeros(a, default_tolerance)
    assert begin != end
    assert begin == 2
    assert end == a.shape[-1]-2
    assert equal_tables(b, e, default_tolerance)

    # Can strip two columns on each side of rank 1 table:
    a = np.ones((5,))
    a[..., 0] = 0.0
    a[..., 1] = 0.0
    a[..., 3] = 0.0
    a[..., 4] = 0.0
    e = np.ones((1,))
    begin, end, b = old_strip_table_zeros(a, default_tolerance)
    assert begin != end
    assert begin == 2
    assert end == a.shape[-1]-2
    assert equal_tables(b, e, default_tolerance)

    # Can strip two columns on each side of rank 3 table:
    a = np.ones((3, 2, 5))
    a[..., 0] = 0.0
    a[..., 1] = 0.0
    a[..., 3] = 0.0
    a[..., 4] = 0.0
    e = np.ones((3, 2, 1))
    begin, end, b = old_strip_table_zeros(a, default_tolerance)
    assert begin != end
    assert begin == 2
    assert end == a.shape[-1]-2
    assert equal_tables(b, e, default_tolerance)

    # Can strip columns on each side and compress the one in the middle:
    a = np.ones((2, 5))
    a[:, 0] = 0.0
    a[:, 1] = 1.0
    a[:, 2] = 0.0
    a[:, 3] = 3.0
    a[:, 4] = 0.0
    e = np.ones((2, 2))
    e[:, 0] = 1.0
    e[:, 1] = 3.0
    dofrange, dofmap, b = strip_table_zeros(a, True, default_tolerance)
    begin, end = dofrange
    assert begin != end
    assert begin == 1
    assert end == a.shape[-1] - 1
    assert list(dofmap) == [1, 3]
    assert equal_tables(b, e, default_tolerance)


def test_unique_tables_some_equal():
    tables = [
        np.zeros((2,)),
        np.ones((2,)),
        np.zeros((3,)),
        np.ones((2,)),
        np.ones((2,))*2,
        np.ones((2,))*2,
        ]
    unique, mapping = build_unique_tables(tables, default_tolerance)
    expected_unique = [
        np.zeros((2,)),
        np.ones((2,)),
        np.zeros((3,)),
        # np.ones((2,)),
        np.ones((2,))*2,
        # np.ones((2,))*2,
        ]
    expected_mapping = dict((i, v) for i, v in enumerate([0, 1, 2, 1, 3, 3]))
    assert mapping == expected_mapping
    assert len(set(itervalues(mapping))) == len(unique)
    for i, t in enumerate(tables):
        assert equal_tables(t, unique[mapping[i]], default_tolerance)


def test_unique_tables_all_equal():
    tables = [np.ones((3, 5))*2.0]*6
    unique, mapping = build_unique_tables(tables, default_tolerance)
    expected_unique = [tables[0]]
    expected_mapping = dict((i, v) for i, v in enumerate([0]*6))
    assert mapping == expected_mapping
    assert len(set(itervalues(mapping))) == len(unique)
    for i, t in enumerate(tables):
        assert equal_tables(t, unique[mapping[i]], default_tolerance)

def test_unique_tables_all_different():
    tables = [
        np.ones((2,)),
        np.ones((2, 3)),
        np.ones((2, 3, 4)),
        np.ones((2, 3, 4, 5)),
        ]
    unique, mapping = build_unique_tables(tables, default_tolerance)
    expected_unique = tables
    expected_mapping = dict((i, i) for i in range(len(tables)))
    assert mapping == expected_mapping
    assert len(set(itervalues(mapping))) == len(unique)
    for i, t in enumerate(tables):
        assert equal_tables(t, unique[mapping[i]], default_tolerance)

def test_unique_tables_string_keys():
    tables = {
        'a': np.zeros((2,)),
        'b': np.ones((2,)),
        'c': np.zeros((3,)),
        'd': np.ones((2,)),
        'e': np.ones((2,))*2,
        'f': np.ones((2,))*2,
        }
    unique, mapping = build_unique_tables(tables, default_tolerance)
    expected_unique = [
        np.zeros((2,)),
        np.ones((2,)),
        np.zeros((3,)),
        # np.ones((2,)),
        np.ones((2,))*2,
        # np.ones((2,))*2,
        ]
    expected_mapping = { 'a': 0, 'b': 1, 'c': 2, 'd': 1, 'e': 3, 'f': 3 }
    assert mapping == expected_mapping
    assert len(set(itervalues(mapping))) == len(unique)
    for i, t in iteritems(tables):
        assert equal_tables(t, unique[mapping[i]], default_tolerance)


def xtest_get_ffc_table_values_scalar_cell():
    cell = triangle
    integral_type = "cell"
    entitytype = "cell"

    class MockElement:
        def value_shape(self): return ()

    element = MockElement()
    component = ()

    avg = None
    entity = 0
    for num_points in (1, 3):
        for num_dofs in (1, 5):
            arr = np.ones((num_dofs, num_points))
            for derivatives in [(0, 0), (1, 1)]:
                # Mocking table built by ffc
                ffc_tables = {
                    num_points: {
                        element: {
                            avg: { # avg
                                entity: { # entityid
                                    derivatives: arr
                                }
                            }
                        }
                    }
                }
                table = get_ffc_table_values(ffc_tables,
                    cell, integral_type,
                    element, avg,
                    entitytype, derivatives, component, 
                    default_tolerance)
                assert equal_tables(table[0, ...], np.transpose(arr), default_tolerance)


def xtest_get_ffc_table_values_vector_facet():
    cell = triangle
    integral_type = "exterior_facet"
    entitytype = "facet"
    num_entities = 3

    class MockElement:
        def value_shape(self): return (2,)

    element = MockElement()
    num_components = 2

    avg = None
    for num_points in (1, 5):
        for num_dofs in (4, 7):
            # Make ones array of the right shape (all dimensions differ to detect algorithm bugs better)
            arr1 = np.ones((num_dofs, num_components, num_points))
            arrays = []
            for i in range(num_entities):
                arr = (i+1.0)*arr1 # Make first digit the facet number (1,2,3)
                for j in range(num_components):
                    arr[:, j,:] += 0.1*(j+1.0) # Make first decimal the component number (1,2)
                for j in range(num_dofs):
                    arr[j,:,:] += 0.01*(j+1.0) # Make second decimal the dof number
                for j in range(num_points):
                    arr[:,:, j] += 0.001*(j+1.0) # Make third decimal the point number
                arrays.append(arr)

            for derivatives in [(), (0,)]:
                # Mocking table built by ffc
                ffc_tables = {
                    num_points: {
                        element: {
                            avg: {
                                entity: { derivatives: arrays[entity] }
                                for entity in range(num_entities)
                            }
                        }
                    }
                }
                # Tables use flattened component, so we can loop over them as integers:
                for component in range(num_components):
                    table = get_ffc_table_values(ffc_tables,
                        cell, integral_type,
                        element, avg,
                        entitytype, derivatives, component,
                        default_tolerance)
                    for i in range(num_entities):
                        # print table[i,...]
                        assert equal_tables(table[i, ...], np.transpose(arrays[i][:, component,:]), default_tolerance)



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2016 Martin Sandve Alnes
#
# This file is part of DOLFIN.
#
# DOLFIN is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DOLFIN is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with DOLFIN. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Marie E. Rognes


class UFCFormNames:

    "Encapsulation of the names related to a generated UFC form."

    def __init__(self, name, coefficient_names, ufc_form_classname,
                 ufc_finite_element_classnames, ufc_dofmap_classnames,
                 superclassname='Form'):
        """Arguments:

        @param name:
            Name of form (e.g. 'a', 'L', 'M').
        @param coefficient_names:
            List of names of form coefficients (e.g. 'f', 'g').
        @param ufc_form_classname:
            Name of ufc::form subclass.
        @param ufc_finite_element_classnames:
            List of names of ufc::finite_element subclasses (length
            rank + num_coefficients).
        @param ufc_dofmap_classnames:
            List of names of ufc::dofmap subclasses (length rank +
            num_coefficients).
        @param superclassname (optional):
            Name of dolfin super class (defaults to 'Form')
        """
        assert len(coefficient_names) <= len(ufc_dofmap_classnames)
        assert len(ufc_finite_element_classnames) == len(ufc_dofmap_classnames)

        self.num_coefficients = len(coefficient_names)
        self.rank = len(ufc_finite_element_classnames) - self.num_coefficients
        self.name = name
        self.coefficient_names = coefficient_names
        self.ufc_form_classname = ufc_form_classname
        self.ufc_finite_element_classnames = ufc_finite_element_classnames
        self.ufc_dofmap_classnames = ufc_dofmap_classnames
        self.superclassname = superclassname

    def __str__(self):
        s = "UFCFormNames instance:\n"
        s += "rank:                      %d\n" % self.rank
        s += "num_coefficients:          %d\n" % self.num_coefficients
        s += "name:                      %s\n" % self.name
        s += "coefficient_names:         %s\n" % str(self.coefficient_names)
        s += "ufc_form_classname:        %s\n" % str(self.ufc_form_classname)
        s += "finite_element_classnames: %s\n" % str(self.ufc_finite_element_classnames)
        s += "ufc_dofmap_classnames:    %s\n" % str(self.ufc_dofmap_classnames)
        return s


class UFCElementNames:

    "Encapsulation of the names related to a generated UFC element."

    def __init__(self, name,
                 ufc_finite_element_classnames,
                 ufc_dofmap_classnames):
        """Arguments:

        """
        assert len(ufc_finite_element_classnames) == len(ufc_dofmap_classnames)

        self.name = name
        self.ufc_finite_element_classnames = ufc_finite_element_classnames
        self.ufc_dofmap_classnames = ufc_dofmap_classnames

    def __str__(self):
        s = "UFCFiniteElementNames instance:\n"
        s += "name:                      %s\n" \
            % self.name
        s += "finite_element_classnames: %s\n"  \
            % str(self.ufc_finite_element_classnames)
        s += "ufc_dofmap_classnames:    %s\n" \
            % str(self.ufc_dofmap_classnames)
        return s



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

from ffc.uflacs.backends.ufc.generator import ufc_generator

### Code generation utilities:

def generate_compute_ATA(L, ATA, A, m, n, index_prefix=""):
    "Generate code to declare and compute ATA[i,j] = sum_k A[k,i]*A[k,j] with given A shaped (m,n)."
    # Loop indices
    i = L.Symbol(index_prefix + "i")
    j = L.Symbol(index_prefix + "j")
    k = L.Symbol(index_prefix + "k")

    # Build A^T*A matrix
    code = [
        L.ArrayDecl("double", ATA, (n, n), values=0),
        L.ForRange(i, 0, n, body=
            L.ForRange(j, 0, n, body=
                L.ForRange(k, 0, m, body=
                    L.AssignAdd(ATA[i, j], A[k, i] * A[k, j])))),
        ]
    return L.StatementList(code)

def generate_accumulation_loop(dst, expr, indices):
    body = L.AssignAdd(dst, expr)
    for index, begin, end in reversed(indices):
        body = L.ForRange(index, begin, end, body=body)
    return body

def cross_expr(a, b):
    def cr(i, j):
        return a[i]*b[j] - a[j]*b[i]
    return [cr(1, 2), cr(2, 0), cr(0, 1)]

def generate_cross_decl(c, a, b):
    return L.ArrayDecl("double", c, values=cross_expr(a, b))

### Inline math expressions:

def det_22(B, i, j, k, l):
    return B[i, k]*B[j, l] - B[i, l]*B[j, k]

def codet_nn(A, rows, cols):
    n = len(rows)
    if n == 2:
        return det_22(A, rows[0], rows[1], cols[0], cols[1])
    else:
        r = rows[0]
        subrows = rows[1:]
        parts = []
        for i, c in enumerate(cols):
            subcols = cols[i+1:] + cols[:i]
            parts.append(A[r, c] * codet_nn(A, subrows, subcols))
        return sum(parts[1:], parts[0])

def det_nn(A, n):
    if n == 1:
        return A[0, 0]
    else:
        ns = list(range(n))
        return codet_nn(A, ns, ns)

def pdet_m1(L, A, m):
    # Special inlined case 1xm for simpler expression
    A2 = A[0,0]*A[0,0]
    for i in range(1, m):
        A2 = A2 + A[i,0]*A[i,0]
    return L.Call("sqrt", A2)

def adj_expr_2x2(A):
    return [[ A[1, 1], -A[0, 1]],
            [-A[1, 0],  A[0, 0]]]

def adj_expr_3x3(A):
    return [[A[2, 2]*A[1, 1] - A[1, 2]*A[2, 1],
             A[0, 2]*A[2, 1] - A[0, 1]*A[2, 2],
             A[0, 1]*A[1, 2] - A[0, 2]*A[1, 1]],
            [A[1, 2]*A[2, 0] - A[2, 2]*A[1, 0],
             A[2, 2]*A[0, 0] - A[0, 2]*A[2, 0],
             A[0, 2]*A[1, 0] - A[1, 2]*A[0, 0]],
            [A[1, 0]*A[2, 1] - A[2, 0]*A[1, 1],
             A[0, 1]*A[2, 0] - A[0, 0]*A[2, 1],
             A[0, 0]*A[1, 1] - A[0, 1]*A[1, 0]]]

def generate_assign_inverse(L, K, J, detJ, gdim, tdim):
    if gdim == tdim:
        if gdim == 1:
            return L.Assign(K[0, 0], L.LiteralFloat(1.0) / J[0, 0])
        elif gdim == 2:
            adj_values = adj_expr_2x2(J)
        elif gdim == 3:
            adj_values = adj_expr_3x3(J)
        else:
            error("Not implemented.")
        return L.StatementList([L.Assign(K[j,i], adj_values[j][i] / detJ)
                                for j in range(tdim) for i in range(gdim)])
    else:
        if tdim == 1:
            # Simpler special case for embedded 1d
            prods = [J[i,0]*J[i,0] for i in range(gdim)]
            s = sum(prods[1:], prods[0])
            return L.StatementList([L.Assign(K[0,i], J[i,0] / s)
                                    for i in range(gdim)])
        else:
            # Generic formulation of Penrose-Moore pseudo-inverse of J: (J.T*J)^-1 * J.T
            i = L.Symbol("i")
            j = L.Symbol("j")
            k = L.Symbol("k")
            JTJ = L.Symbol("JTJ")
            JTJf = L.FlattenedArray(JTJ, dims=(tdim, tdim))
            detJTJ = detJ*detJ
            JTJinv = L.Symbol("JTJinv")
            JTJinvf = L.FlattenedArray(JTJinv, dims=(tdim, tdim))
            code = [
                L.Comment("Compute J^T J"),
                L.ArrayDecl("double", JTJ, (tdim*tdim,), values=0),
                L.ForRange(k, 0, tdim, body=
                    L.ForRange(j, 0, tdim, body=
                        L.ForRange(i, 0, gdim, body=
                            L.AssignAdd(JTJf[k, j], J[i, k] * J[i, j])))),
                L.Comment("Compute inverse(J^T J)"),
                L.ArrayDecl("double", JTJinv, (tdim*tdim,)),
                generate_assign_inverse(L, JTJinvf, JTJf, detJTJ, tdim, tdim),
                L.Comment("Compute K = inverse(J^T J) * J"),
                L.ForRange(k, 0, tdim, body=
                    L.ForRange(i, 0, gdim, body=
                            L.Assign(K[k, i], L.LiteralFloat(0.0)))),
                L.ForRange(k, 0, tdim, body=
                    L.ForRange(i, 0, gdim, body=
                        L.ForRange(j, 0, tdim, body=
                            L.AssignAdd(K[k, i], JTJinvf[k, j] * J[i, j])))),
                ]
            return L.StatementList(code)


class ufc_coordinate_mapping(ufc_generator):
    def __init__(self):
        ufc_generator.__init__(self, "coordinate_mapping")

    def cell_shape(self, L, ir):
        name = ir["cell_shape"]
        return L.Return(L.Symbol(name))

    def topological_dimension(self, L, ir):
        "Default implementation of returning topological dimension fetched from ir."
        value = ir["topological_dimension"]
        return L.Return(L.LiteralInt(value))

    def geometric_dimension(self, L, ir):
        "Default implementation of returning geometric dimension fetched from ir."
        value = ir["geometric_dimension"]
        return L.Return(L.LiteralInt(value))

    def create_coordinate_finite_element(self, L, ir):
        classname = ir["create_coordinate_finite_element"]
        return L.Return(L.New(classname))

    def create_coordinate_dofmap(self, L, ir):
        classname = ir["create_coordinate_dofmap"]
        return L.Return(L.New(classname))

    def compute_physical_coordinates(self, L, ir): # FIXME: Implement evaluate_reference_basis
        num_dofs = ir["num_scalar_coordinate_element_dofs"]
        scalar_coordinate_element_classname = ir["scalar_coordinate_finite_element_classname"]

        # Dimensions
        gdim = ir["geometric_dimension"]
        tdim = ir["topological_dimension"]
        num_points = L.Symbol("num_points")

        # Loop indices
        ip = L.Symbol("ip")
        i = L.Symbol("i")
        j = L.Symbol("j")
        d = L.Symbol("d")

        # Input cell data
        coordinate_dofs = L.FlattenedArray(L.Symbol("coordinate_dofs"), dims=(num_dofs, gdim))

        # Output geometry
        x = L.FlattenedArray(L.Symbol("x"), dims=(num_points, gdim))

        # Input geometry
        X = L.FlattenedArray(L.Symbol("X"), dims=(num_points, tdim))

        # Computing table one point at a time instead of using
        # num_points will allow skipping dynamic allocation
        one_point = 1

        # FIXME: Use correct approach and function name to compute basis values here
        if 1:
            # Call static function on element class
            define_element = []
            func = "%s::_evaluate_reference_basis" % (scalar_coordinate_element_classname,)
        else:
            # Define scalar finite element instance (stateless, so placing this on the stack is free)
            define_element = [L.VariableDecl(scalar_coordinate_element_classname, "element")]
            func = "element.evaluate_reference_basis"

        # Symbols for local basis values table
        phi_sym = L.Symbol("phi")
        phi = L.FlattenedArray(phi_sym, dims=(num_dofs,)) # FIXME: Match evaluate_reference_basis array layout

        # Arguments to evaluate_reference_basis
        args = (phi_sym, one_point, L.AddressOf(X[ip, 0])) # FIXME: Match evaluate_reference_basis signature

        # For each point, compute basis values and accumulate into the right x
        code = L.StatementList(define_element + [
            L.ArrayDecl("double", phi_sym, (one_point*num_dofs,)),
            L.ForRange(ip, 0, num_points, body=L.StatementList([
                L.Comment("Compute basis values of coordinate element"),
                L.Call(func, args),
                L.Comment("Compute x"),
                L.ForRange(i, 0, gdim, body=
                    L.ForRange(d, 0, num_dofs, body=
                        L.AssignAdd(x[ip, i], coordinate_dofs[d, i]*phi[d]))),
                ]))
            ])
        return code

    def compute_reference_coordinates(self, L, ir):
        degree = ir["coordinate_element_degree"]
        if degree == 1:
            # Special case optimized for affine mesh (possibly room for further optimization)
            return self._compute_reference_coordinates_affine(L, ir)
        else:
            # General case with newton loop to solve F(X) = x(X) - x0 = 0
            return self._compute_reference_coordinates_newton(L, ir)

    def _compute_reference_coordinates_affine(self, L, ir): # TODO: Test!
        # Dimensions
        gdim = ir["geometric_dimension"]
        tdim = ir["topological_dimension"]
        cellname = ir["cell_shape"]
        num_points = L.Symbol("num_points")

        # Number of dofs for a scalar component
        num_dofs = ir["num_scalar_coordinate_element_dofs"]

        # Loop indices
        ip = L.Symbol("ip") # point
        i = L.Symbol("i")   # gdim
        j = L.Symbol("j")   # tdim
        k = L.Symbol("k")   # sum iteration

        # Output geometry
        X = L.FlattenedArray(L.Symbol("X"), dims=(num_points, tdim))

        # Input geometry
        x = L.FlattenedArray(L.Symbol("x"), dims=(num_points, gdim))

        # Input cell data
        coordinate_dofs = L.FlattenedArray(L.Symbol("coordinate_dofs"), dims=(num_dofs, gdim))

        # Tables of coordinate basis function values and derivatives at
        # X=0 and X=midpoint available through ir. This is useful in
        # several geometry functions.
        tables = ir["tables"]

        # Check the table shapes against our expectations
        x_table = tables["x0"]
        J_table = tables["J0"]
        assert x_table.shape == (num_dofs,)
        assert J_table.shape == (tdim, num_dofs)

        # Table symbols
        phi_X0 = L.Symbol("phi_X0")
        dphi_X0 = L.Symbol("dphi_X0")

        # Table declarations
        table_decls = [
            L.ArrayDecl("static const double", phi_X0, sizes=x_table.shape, values=tables["x0"]),
            L.ArrayDecl("static const double", dphi_X0, sizes=J_table.shape, values=tables["J0"]),
            ]

        # Compute x0 = x(X=0) (optimized by precomputing basis at X=0)
        x0 = L.Symbol("x0")
        compute_x0 = [
            L.ArrayDecl("double", x0, sizes=(gdim,), values=0),
            L.ForRange(i, 0, gdim, body=
                L.ForRange(k, 0, num_dofs, body=
                    L.AssignAdd(x0[i], coordinate_dofs[k, i] * phi_X0[k]))),
            ]

        # Compute J0 = J(X=0) (optimized by precomputing basis at X=0)
        J0 = L.Symbol("J0")
        compute_J0 = [
            L.ArrayDecl("double", J0, sizes=(gdim*tdim,), values=0),
            L.ForRange(i, 0, gdim, body=
                L.ForRange(j, 0, tdim, body=
                    L.ForRange(k, 0, num_dofs, body=
                        L.AssignAdd(J0[i*tdim + j], coordinate_dofs[k, i] * dphi_X0[j, k])))),
            ]

        # Compute K0 = inv(J0) (and intermediate value det(J0))
        detJ0 = L.Symbol("detJ0")
        K0 = L.Symbol("K0")
        compute_K0 = [
            L.ArrayDecl("double", detJ0, sizes=(1,)),
            L.Call("compute_jacobian_determinants", (detJ0, 1, J0)),
            L.ArrayDecl("double", K0, sizes=(tdim*gdim,)),
            L.Call("compute_jacobian_inverses", (K0, 1, J0, detJ0)),
            ]

        # Compute X = K0*(x-x0) for each physical point x
        compute_X = [
            L.ForRange(ip, 0, num_points, body=
                L.ForRange(j, 0, tdim, body=
                    L.ForRange(i, 0, gdim, body=
                        L.AssignAdd(X[ip, j], K0[j, i]*(x[ip, i] - x0[i]))))),
            ]

        # Stitch it together
        code = L.StatementList(table_decls + compute_x0 + compute_J0 + compute_K0 + compute_X)
        return code

    def _compute_reference_coordinates_newton(self, L, ir): # TODO: Test! TODO: Determine stopping criteria to use.
        """Solves x(X) = x0 for X.

        Find X such that, given x0,

          F(X) = x(X) - x0 = 0

        Newton iteration is:

          X^0 = midpoint of reference cell
          until convergence:
              dF/dX(X^k) dX^k = -F(X^k)
              X^{k+1} = X^k + dX^k

        The Jacobian is just the usual Jacobian of the geometry mapping:

            dF/dX = dx/dX = J

        and its inverse is the usual K = J^-1.

        Because J is small, we can invert it directly and rewrite

            Jk dX = -(xk - x0) = x0 - xk

        into

            Xk = midpoint of reference cell
            for k in range(maxiter):
                xk = x(Xk)
                Jk = J(Xk)
                Kk = inverse(Jk)
                dX = Kk (x0 - xk)
                Xk += dX
                if dX sufficiently small: break
        """
        # Dimensions
        gdim = ir["geometric_dimension"]
        tdim = ir["topological_dimension"]
        cellname = ir["cell_shape"]
        num_points = L.Symbol("num_points")

        degree = ir["coordinate_element_degree"]

        # Computing table one point at a time instead of vectorized
        # over num_points will allow skipping dynamic allocation
        one_point = 1

        # Loop indices
        ip = L.Symbol("ip") # point
        i = L.Symbol("i")   # gdim
        j = L.Symbol("j")   # tdim
        k = L.Symbol("k")   # iteration

        # Input cell data
        coordinate_dofs = L.Symbol("coordinate_dofs")
        cell_orientation = L.Symbol("cell_orientation")

        # Output geometry
        X = L.FlattenedArray(L.Symbol("X"), dims=(num_points, tdim))

        # Input geometry
        x = L.FlattenedArray(L.Symbol("x"), dims=(num_points, gdim))

        # Find X=Xk such that xk(Xk) = x or F(Xk) = xk(Xk) - x = 0.
        # Newtons method is then:
        #   X0 = cell midpoint
        #   dF/dX dX = -F
        #   Xk = Xk + dX
        # Note that
        #   dF/dX = dx/dX = J
        # giving
        #   inverse(dF/dX) = K
        # such that dX = -K*(xk(Xk) - x)

        # Newtons method is then:
        # for each ip:
        #   xgoal = x[ip]
        #   X0 = cell midpoint
        #   until convergence:
        #       K = K(Xk)
        #       dX = K*(xk(Xk) - x)
        #       Xk -= dX
        #   X[ip] = Xk

        # Symbols for arrays used below
        Xk = L.Symbol("Xk")
        xgoal = L.Symbol("xgoal")
        dX = L.Symbol("dX")
        xk = L.Symbol("xk")
        J = L.Symbol("J")
        detJ = L.Symbol("detJ")
        K = L.Symbol("K")

        # Symbol for ufc_geometry cell midpoint definition
        mp = L.Symbol("%s_midpoint" % cellname)

        # Variables for stopping criteria
        # TODO: Check if these are good convergence criteria, e.g. is epsilon=1e-6 and iterations=degree sufficient?
        max_iter = L.LiteralInt(2*degree)
        epsilon = L.LiteralFloat(1e-14)
        # TODO: Could also easily make criteria input if desired
        #max_iter = L.Symbol("iterations")
        #epsilon = L.Symbol("epsilon")
        dX2 = L.Symbol("dX2")

        # Wrap K as flattened array for convenient indexing Kf[j,i]
        Kf = L.FlattenedArray(K, dims=(tdim, gdim))

        # declare xgoal = x[ip]
        # declare Xk = initial value
        newton_init = [
            L.Comment("Declare xgoal to hold the current x coordinate value"),
            L.ArrayDecl("const double", xgoal, (gdim,), values=[x[ip][iv] for iv in range(gdim)]),
            L.Comment("Declare Xk iterate with initial value equal to reference cell midpoint"),
            L.ArrayDecl("double", Xk, (tdim,), values=[mp[c] for c in range(tdim)]),
            ]

        # TODO: Use ir["tables"]["xm|Jm"] to compute x(Xm) and J(Xm) for an optimized first iteration

        part1 = [
            L.Comment("Declare intermediate arrays to hold results of compute_geometry call"),
            L.ArrayDecl("double", xk, (gdim,)),
            L.ArrayDecl("double", J, (gdim*tdim,)),
            L.ArrayDecl("double", detJ, (1,)),
            L.ArrayDecl("double", K, (tdim*gdim,)),

            L.Comment("Compute K = J^-1 for one point, (J and detJ are only used as"),
            L.Comment("intermediate storage inside compute_geometry, not used out here"),
            L.Call("compute_geometry",
                   (xk, J, detJ, K, one_point,
                    Xk, coordinate_dofs, cell_orientation)),
            ]

        # Newton body with stopping criteria |dX|^2 < epsilon
        newton_body = part1 + [
            L.Comment("Declare dX increment to be computed, initialized to zero"),
            L.ArrayDecl("double", dX, (tdim,), values=0.0),

            L.Comment("Compute dX[j] = sum_i K_ji * (x_i - x(Xk)_i)"),
            L.ForRange(j, 0, tdim, body=
                       L.ForRange(i, 0, gdim, body=
                                  L.AssignAdd(dX[j], Kf[j,i]*(xgoal[i] - xk[i])))),

            L.Comment("Compute |dX|^2"),
            L.VariableDecl("double", dX2, value=0.0),
            L.ForRange(j, 0, tdim, body=L.AssignAdd(dX2, dX[j]*dX[j])),

            L.Comment("Break if converged (before X += dX such that X,J,detJ,K are consistent)"),
            L.If(L.LT(dX2, epsilon), L.Break()),

            L.Comment("Update Xk += dX"),
            L.ForRange(j, 0, tdim, body=L.AssignAdd(Xk[j], dX[j])),
            ]

        # Use this instead for fixed iteration number
        alternative_newton_body = part1 + [
            L.Comment("Compute Xk[j] += sum_i K_ji * (x_i - x(Xk)_i)"),
            L.ForRange(j, 0, tdim, body=
                       L.ForRange(i, 0, gdim, body=
                                  L.AssignAdd(Xk[j], Kf[j,i]*(xgoal[i] - xk[i])))),
            ]

        # Loop until convergence
        newton_loop = L.ForRange(k, 0, max_iter, body=newton_body)

        # X[ip] = Xk
        newton_finish = L.ForRange(j, 0, tdim, body=L.Assign(X[ip][j], Xk[j]))

        # Carry out newton loop for each point
        code = L.ForRange(ip, 0, num_points,
                          body=L.StatementList([newton_init, newton_loop, newton_finish]))
        return code

    def compute_jacobians(self, L, ir): # FIXME: Implement evaluate_reference_basis_derivatives
        num_dofs = ir["num_scalar_coordinate_element_dofs"]
        scalar_coordinate_element_classname = ir["scalar_coordinate_finite_element_classname"]

        # Dimensions
        gdim = ir["geometric_dimension"]
        tdim = ir["topological_dimension"]
        num_points = L.Symbol("num_points")

        # Loop indices
        ip = L.Symbol("ip")
        i = L.Symbol("i")
        j = L.Symbol("j")
        d = L.Symbol("d")

        # Input cell data
        coordinate_dofs = L.FlattenedArray(L.Symbol("coordinate_dofs"), dims=(num_dofs, gdim))

        # Output geometry
        J = L.FlattenedArray(L.Symbol("J"), dims=(num_points, gdim, tdim))

        # Input geometry
        X = L.FlattenedArray(L.Symbol("X"), dims=(num_points, tdim))

        # Computing table one point at a time instead of using
        # num_points will allow skipping dynamic allocation
        one_point = 1

        # FIXME: Use correct approach and function name to compute basis derivatives here
        if 1:
            # Call static function on element class
            define_element = []
            func = "%s::_evaluate_reference_basis_derivatives" % (scalar_coordinate_element_classname,)
        else:
            # Define scalar finite element instance (stateless, so placing this on the stack is free)
            define_element = [L.VariableDecl(scalar_coordinate_element_classname, "element")]
            func = "element.evaluate_reference_basis_derivatives"

        # Symbols for local basis derivatives table
        dphi_sym = L.Symbol("dphi")
        dphi = L.FlattenedArray(dphi_sym, dims=(tdim, num_dofs)) # FIXME: Match array layout of evaluate_reference_basis_derivatives

        # Arguments to evaluate_reference_basis_derivatives
        args = (dphi_sym, one_point, L.AddressOf(X[ip, 0])) # FIXME: Match evaluate_reference_basis_derivatives signature

        # For each point, compute basis derivatives and accumulate into the right J
        code = L.StatementList(define_element + [
            L.ArrayDecl("double", dphi_sym, (one_point*tdim*num_dofs,)),
            L.ForRange(ip, 0, num_points, body=L.StatementList([
                L.Comment("Compute basis derivatives of coordinate element"),
                L.Call(func, args),
                L.Comment("Compute J"),
                L.ForRange(i, 0, gdim, body=
                    L.ForRange(j, 0, tdim, body=
                        L.ForRange(d, 0, num_dofs, body=
                            L.AssignAdd(J[ip, i, j], coordinate_dofs[d, i]*dphi[j, d]))))
                ]))
            ])
        return code

    def compute_jacobian_determinants(self, L, ir): # Looks good on inspection. TODO: Test!
        # Dimensions
        gdim = ir["geometric_dimension"]
        tdim = ir["topological_dimension"]
        num_points = L.Symbol("num_points")

        # Loop indices
        ip = L.Symbol("ip")

        # Output geometry
        detJ = L.Symbol("detJ")[ip]

        # Input geometry
        J = L.FlattenedArray(L.Symbol("J"), dims=(num_points, gdim, tdim))
        cell_orientation = L.Symbol("cell_orientation")

        # Assign det expression to detJ # TODO: Call Eigen instead?
        if gdim == tdim:
            body = L.Assign(detJ, det_nn(J[ip], gdim))
        elif tdim == 1:
            body = L.Assign(detJ, cell_orientation*pdet_m1(A, gdim))
        #elif tdim == 2 and gdim == 3:
        #    body = L.Assign(detJ, cell_orientation*pdet_32(A)) # Possible optimization not implemented here
        else:
            JTJ = L.Symbol("JTJ")
            body = L.StatementList([
                generate_compute_ATA(L, JTJ, J[ip], gdim, tdim),
                L.Assign(detJ, cell_orientation*L.Call("sqrt", det_nn(JTJ, tdim))),
                ])

        # Carry out for all points
        loop = L.ForRange(ip, 0, num_points, body=body)

        code = loop #[defines, loop]
        return code

    def compute_jacobian_inverses(self, L, ir): # TODO: Test! Cover all combinations of gdim,tdim=1,2,3!
        # Dimensions
        gdim = ir["geometric_dimension"]
        tdim = ir["topological_dimension"]
        num_points = L.Symbol("num_points")

        # Loop indices
        ip = L.Symbol("ip")

        # Output geometry
        K = L.FlattenedArray(L.Symbol("K"), dims=(num_points, tdim, gdim))

        # Input geometry
        J = L.FlattenedArray(L.Symbol("J"), dims=(num_points, gdim, tdim))
        detJ = L.Symbol("detJ")

        # Assign to K[j][i] for each component j,i
        body = generate_assign_inverse(L, K[ip], J[ip], detJ[ip], gdim, tdim)

        # Carry out for all points
        return L.ForRange(ip, 0, num_points, body=body)

    def compute_geometry(self, L, ir): # Output looks good on inspection. TODO: Test!
        # Output geometry
        x = L.Symbol("x")
        J = L.Symbol("J")
        detJ = L.Symbol("detJ")
        K = L.Symbol("K")

        # Dimensions
        num_points = L.Symbol("num_points")

        # Input geometry
        X = L.Symbol("X")

        # Input cell data
        coordinate_dofs = L.Symbol("coordinate_dofs")
        cell_orientation = L.Symbol("cell_orientation")

        # All arguments
        args = (x, J, detJ, K, num_points, X, coordinate_dofs, cell_orientation)

        # Just chain calls to other functions here
        code = [
            L.Call("compute_physical_coordinates", (x, num_points, X, coordinate_dofs, cell_orientation)),
            L.Call("compute_jacobians", (J, num_points, X, coordinate_dofs, cell_orientation)),
            L.Call("compute_jacobian_determinants", (detJ, num_points, J)),
            L.Call("compute_jacobian_inverses", (K, num_points, J, detJ)),
            ]
        return L.StatementList(code)



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testElasWeighted():
    expr = Product([
        Symbol('W4', IP),
        Sum([
            Product([
                Symbol('FE0_C1_D01_ip_j', BASIS),
                Symbol('FE0_C1_D01_ip_k', BASIS),
                Symbol('Jinv_00', GEO),
                Symbol('w1', GEO)
            ]),
            Product([
                Symbol('FE0_C1_D01_ip_j', BASIS),
                Symbol('FE0_C1_D01_ip_k', BASIS),
                Symbol('Jinv_01', GEO),
                Symbol('w0', GEO)
            ]),
            Product([
                Symbol('w2', GEO),
                Sum([
                    Product([
                        Symbol('FE0_C1_D01_ip_j', BASIS),
                        Symbol('FE0_C1_D01_ip_k', BASIS),
                        Symbol('Jinv_00', GEO),
                        Symbol('w1', GEO)
                    ]),
                    Product([
                        Symbol('FE0_C1_D01_ip_j', BASIS),
                        Symbol('FE0_C1_D01_ip_k', BASIS),
                        Symbol('Jinv_01', GEO),
                        Symbol('w0', GEO)
                    ])
                ])
            ])
        ])
    ])

    expr_exp = expr.expand()
    expr_red = expr_exp.reduce_ops()

    det, W4, w0, w1, w2, Jinv_00, Jinv_01, Jinv_11, Jinv_10, FE0_C1_D01_ip_j, FE0_C1_D01_ip_k = [0.123 + i for i in range(11)]
    assert round(eval(str(expr)) - eval(str(expr_exp))) == 0.0
    assert round(eval(str(expr)) - eval(str(expr_red))) == 0.0
    assert expr.ops() == 17
    assert expr_exp.ops() == 21
    assert expr_red.ops() == 10

    # Generate code
    ip_consts = {}
    geo_consts = {}
    trans_set = set()

    start = time.time()
    opt_code = optimise_code(expr, ip_consts, geo_consts, trans_set)

    G = [eval(str(list(geo_consts.items())[0][0]))]
    I = [eval(str(list(ip_consts.items())[0][0]))]
    assert round(eval(str(expr)) - eval(str(opt_code))) == 0.0
